import unittest
import yggdrasil.drivers.tests.test_ExecutableModelDriver as parent


class TestOSRModelParam(parent.TestExecutableModelParam):
    r"""Test parameters for OSRModelDriver."""

    driver = "OSRModelDriver"

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for creating a class instance."""
        out = super(TestOSRModelParam, self).inst_kwargs
        out.update(timesync={'name': 'timesync',
                             'inputs': ['carbonAllocation2Roots'],
                             'outputs': ['carbonAllocation2Roots']},
                   copy_xml_to_osr=True)
        return out

    @unittest.skipIf(True, "Requires partner model.")
    def test_run_model(self, *args, **kwargs):  # pragma: no cover
        r"""Test running script used without debug."""
        pass

    @unittest.skipIf(True, "Requires partner model.")
    def test_valgrind(self, *args, **kwargs):  # pragma: no cover
        pass

    @unittest.skipIf(True, "Requires partner model.")
    def test_strace(self, *args, **kwargs):  # pragma: no cover
        pass


class TestOSRModelDriverNoInit(TestOSRModelParam,
                               parent.TestExecutableModelDriverNoInit):
    r"""Test runner for OSRModelDriver without init."""

    def test_cleanup_dependencies(self):
        r"""Test cleanup_dependencies method."""
        super(TestOSRModelDriverNoInit, self).test_cleanup_dependencies()
        self.import_cls.cleanup_dependencies()
        

class TestOSRModelDriverNoStart(TestOSRModelParam,
                                parent.TestExecutableModelDriverNoStart):
    r"""Test runner for OSRModelDriver without start."""
    pass


class TestOSRModelDriver(TestOSRModelParam,
                         parent.TestExecutableModelDriver):
    r"""Test runner for OSRModelDriver."""
    pass



import os
from yggdrasil.demos.tests import DemoTstBase, _demo_dir
import unittest
try:
    import trimesh
except ImportError:  # pragma: debug
    trimesh = None


@unittest.skipIf(trimesh is None, "Trimesh is not installed")
class TestCiS2021HackathonDemo(DemoTstBase):

    demo_name = 'CiS2021-hackathon'
    runs = {'light_v0': ('yamls/light_v0_python.yml', 'yamls/connections_v0.yml'),
            'shoot_v0': ('yamls/shoot_v0.yml', ),
            'shoot_v1': ('yamls/shoot_v1.yml', 'yamls/light_v0_python.yml',
                         'yamls/connections_v1.yml'),
            'shoot_v2': ('yamls/light_v1_python.yml', 'yamls/shoot_v2.yml'),
            'shoot_v2_split': ('yamls/light_v1_python.yml',
                               'yamls/shoot_v2_split.yml'),
            'shoot_v2_copies': ('yamls/light_v2_python.yml',
                                'yamls/shoot_v2_copies.yml'),
            'roots_v0': ('yamls/roots_v0.yml', ),
            'timesync': ('yamls/roots_v1.yml', 'yamls/shoot_v3.yml',
                         'yamls/timesync.yml', 'yamls/light_v1_python.yml')}

    def __init__(self, *args, **kwargs):
        super(TestCiS2021HackathonDemo, self).__init__(*args, **kwargs)
        out_dir = os.path.join(_demo_dir, self.demo_name, 'output')
        if not os.path.isdir(out_dir):
            os.mkdir(out_dir)



import os
import shutil
import yggdrasil.drivers.tests.test_CompiledModelDriver as parent


class TestBuildModelParam(parent.TestCompiledModelParam):
    r"""Test parameters for BuildModelDriver."""
    pass


class TestBuildModelDriverNoInit(TestBuildModelParam,
                                 parent.TestCompiledModelDriverNoInit):
    """Test runner for BuildModelDriver without creating an instance."""

    test_build = None
    test_get_tool = None
    test_get_dependency_info = None
    test_get_dependency_source = None
    test_get_dependency_object = None
    test_get_dependency_library = None
    test_get_dependency_include_dirs = None
    test_get_dependency_order = None
    test_invalid_function_param = None


class TestBuildModelDriverNoStart(TestBuildModelParam,
                                  parent.TestCompiledModelDriverNoStart):
    r"""Test runner for BuildModelDriver without start."""

    test_compilers = None

    def test_get_language_for_source(self):
        r"""Test the get_language_for_source method."""
        buildfile = None
        if self.import_cls.buildfile_base:
            buildfile = os.path.join(os.path.dirname(self.src[0]),
                                     self.import_cls.buildfile_base)
            buildfile_cache = '_copy'.join(os.path.splitext(buildfile))
        self.import_cls.get_language_for_source(self.src)
        self.import_cls.get_language_for_source(os.path.dirname(self.src[0]))
        try:
            if buildfile:
                shutil.move(buildfile, buildfile_cache)
            self.import_cls.get_language_for_source(self.src[0])
        finally:
            if buildfile:
                shutil.move(buildfile_cache, buildfile)


class TestBuildModelDriver(TestBuildModelParam,
                           parent.TestCompiledModelDriver):
    r"""Test runner for BuildModelDriver."""
    pass



import os
import yggdrasil.drivers.tests.test_CompiledModelDriver as parent
from yggdrasil.drivers.FortranModelDriver import FortranModelDriver


class TestFortranModelParam(parent.TestCompiledModelParam):
    r"""Test parameters for FortranModelDriver."""

    driver = 'FortranModelDriver'
    
    def __init__(self, *args, **kwargs):
        super(TestFortranModelParam, self).__init__(*args, **kwargs)
        script_dir = os.path.dirname(self.src[0])
        self.args = [self.args[0], '1']
        if FortranModelDriver.is_installed():
            compiler = FortranModelDriver.get_tool('compiler')
            linker = FortranModelDriver.get_tool('linker')
            include_flag = compiler.create_flag('include_dirs', script_dir)
            library_flag = linker.create_flag('library_dirs', script_dir)
            self._inst_kwargs.update(compiler_flags=include_flag,
                                     linker_flags=library_flag)


class TestFortranModelDriverNoInit(TestFortranModelParam,
                                   parent.TestCompiledModelDriverNoInit):
    r"""Test runner for FortranModelDriver without init."""

    def get_test_types(self):
        r"""Return the list of tuples mapping json type to expected native type."""
        out = super(TestFortranModelDriverNoInit, self).get_test_types()
        for i, (k, v) in enumerate(out):
            knew = k
            vnew = v
            if vnew == '*':
                knew = {'type': knew, 'subtype': 'float',
                        'precision': 32}
                vnew = 'real(kind = 4)'
                if k == '1darray':
                    knew['length'] = 3
                    vnew += ', dimension(3)'
                elif k == 'ndarray':
                    knew['shape'] = (3, 4)
                    vnew += ', dimension(3,4)'
            elif 'X' in vnew:
                if vnew.startswith('complex'):
                    knew = {'type': knew, 'precision': 128}
                elif 'ISO_10646' in vnew:
                    knew = {'type': knew, 'precision': 4 * 64}
                else:
                    knew = {'type': knew, 'precision': 64}
                vnew = vnew.replace('X', '8')
            if vnew.startswith('ygg'):
                vnew = 'type(%s)' % vnew
            out[i] = (knew, vnew)
                
        return out
    
    def test_write_function_def(self, *args, **kwargs):
        r"""Test writing and running a function definition."""
        kwargs['declare_functions_as_var'] = True
        super(TestFortranModelDriverNoInit, self).test_write_function_def(*args, **kwargs)
        
    def test_write_function_def_single(self):
        r"""Test writing and running a function definition with single output."""
        inputs = [{'name': 'x', 'value': 1.0,
                   'datatype': {'type': 'float',
                                'precision': 32,
                                'units': 'cm'}}]
        outputs = [{'name': 'y',
                    'datatype': {'type': 'float',
                                 'precision': 32,
                                 'units': 'cm'}}]
        self.test_write_function_def(inputs=inputs, outputs=outputs,
                                     outputs_in_inputs=False,
                                     dont_add_lengths=True)
        
    def test_write_function_def_void(self):
        r"""Test writing and running a function definition with no output."""
        inputs = [{'name': 'x', 'value': 1.0,
                   'datatype': {'type': 'float',
                                'precision': 32,
                                'units': 'cm'}}]
        outputs = []
        self.test_write_function_def(inputs=inputs, outputs=outputs,
                                     outputs_in_inputs=False)
        
    def test_write_function_def_string(self):
        r"""Test writing and running a function definition with no length var."""
        inputs = [{'name': 'x', 'value': '"hello"',
                   'length_var': 'length_x',
                   'datatype': {'type': 'string',
                                'precision': 20,
                                'units': ''}},
                  {'name': 'length_x', 'value': 5,
                   'datatype': {'type': 'uint',
                                'precision': 64},
                   'is_length_var': True}]
        outputs = [{'name': 'y',
                    'length_var': 'length_y',
                    'datatype': {'type': 'string',
                                 'precision': 20,
                                 'units': ''}},
                   {'name': 'length_y',
                    'datatype': {'type': 'uint',
                                 'precision': 64},
                    'is_length_var': True}]
        self.test_write_function_def(inputs=inputs, outputs=outputs,
                                     dont_add_lengths=True)

    def test_write_function_def_return(self):
        r"""Test writing and running a function definition that returns output
        (rather than passing output but parameter reference)."""
        inputs = [{'name': 'x', 'value': 1.0,
                   'datatype': {'type': 'float',
                                'precision': 32,
                                'units': 'cm'}}]
        outputs = [{'name': 'y',
                    'datatype': {'type': 'float',
                                 'precision': 32,
                                 'units': 'cm'}}]
        self.test_write_function_def(inputs=inputs, outputs=outputs,
                                     outputs_in_inputs=False,
                                     guess_at_outputs_in_inputs=True)

    def test_write_function_def_guess(self):
        r"""Test writing and running a function definition where outputs
        are determined from the intents."""
        inputs = [{'name': 'x', 'value': 1.0,
                   'datatype': {'type': 'float',
                                'precision': 32,
                                'units': 'cm'}}]
        outputs = [{'name': 'y',
                    'datatype': {'type': 'float',
                                 'precision': 32,
                                 'units': 'cm'}}]
        self.test_write_function_def(inputs=inputs, outputs=outputs,
                                     guess_at_outputs_in_inputs=True)

    def test_write_try_except(self, **kwargs):
        r"""Test writing a try/except block."""
        pass

    def test_split_line(self, vals=None):
        r"""Test split_line."""
        if vals is None:
            vals = [('abcdef', {'length': 3, 'force_split': True},
                     ['ab&', '     &cdef']),
                    ('    abc', {'length': 3, 'force_split': True},
                     ['    abc'])]
        super(TestFortranModelDriverNoInit, self).test_split_line(vals=vals)

            
class TestFortranModelDriverNoStart(TestFortranModelParam,
                                    parent.TestCompiledModelDriverNoStart):
    r"""Test runner for FortranModelDriver without start."""

    def test_parse_arguments(self):
        r"""Run test to initialize driver using the executable."""
        x = os.path.splitext(self.instance.source_files[0])[0] + '.out'
        new_inst = self.import_cls('test_name', [x], skip_compile=True)
        self.assert_equal(new_inst.model_file, x)
        self.assert_equal(new_inst.source_files, self.instance.source_files[:1])
        

class TestFortranModelDriver(TestFortranModelParam,
                             parent.TestCompiledModelDriver):
    r"""Test runner for FortranModelDriver."""
    pass



import os
import trimesh
import argparse

_dir = os.path.dirname(os.path.realpath(__file__))

# Parse command-line arguments
parser = argparse.ArgumentParser("Simulate a shoot's growth over time.")
parser.add_argument('tmin', help='Starting time (in hours)', type=float)
parser.add_argument('tmax', help='Ending time (in hours)', type=float)
parser.add_argument('tstep', help='Time step (in hours)', type=float)
parser.add_argument('--meshfile', help='Path to file where mesh is stored.',
                    default='../meshes/plants-2.obj')
args = parser.parse_args()
tmin = args.tmin
tmax = args.tmax
tstep = args.tstep
mesh = trimesh.load_mesh(args.meshfile)

# Set initial conditions
mass = 2000.0
t = tmin
i = 0

# Check if model is running as a part of an yggdrasil integration
with_yggdrasil = os.environ.get('YGG_SUBPROCESS', False)

# If the model is running as part of an yggdrasil integration, import
# the relevant yggdrasil routines and use the interface routine to
# complete the connection defined in the YAML
if with_yggdrasil:
    from yggdrasil import units
    from yggdrasil.languages.Python.YggInterface import YggRpcClient
    light_rpc = YggRpcClient('light_shoot')

# Continue simulation until time limit is reached
while t <= tmax:

    # If running as part an yggdrasil integration, send the time and
    # maximum height of the mesh to the height channel with units
    if with_yggdrasil:
        flag, intensity = light_rpc.call(
            [units.add_units(t, 'hrs'),
             units.add_units(max(mesh.vertices[:, 2]), 'm')])
        if not flag:
            raise Exception("Error calling the light model.")

        # Compute the scale factor using intensity, stripping units
        # of the result to allow use with trimesh
        # (pretend this is a biologically complex calculation)
        scale = units.get_data(
            units.add_units(mass, 'g') * intensity /
            units.add_units(4.0e10, 'g*erg/(cm**2*s)'))
        
    else:
        # Compute the scale factor
        # (pretend this is a biologically complex calculation)
        scale = mass / 4.5e4

    # Grow the shoot
    # (pretend this is a biologically complex calculation)
    mesh.vertices[:, 2] += mesh.vertices[:, 2] * scale
    mass += mass * scale

    # Save mesh for this timestep
    filename_mesh = os.path.join(_dir, f'../output/mesh_{i:03d}.obj')
    with open(filename_mesh, 'w') as fd:
        mesh.export(fd, 'obj')

    # Advance time step
    t += tstep
    i += 1



import uuid
from collections import OrderedDict
from yggdrasil.communication import CommBase, new_comm, get_comm, import_comm


class ClientComm(CommBase.CommBase):
    r"""Class for handling Client side communication.

    Args:
        name (str): The environment variable where communication address is
            stored.
        request_commtype (str, optional): Comm class that should be used for
            the request comm. Defaults to None.
        response_kwargs (dict, optional): Keyword arguments for the response
            comm. Defaults to empty dict.
        direct_connection (bool, optional): If True, the comm will be
            directly connected to a ServerComm. Defaults to False.
        **kwargs: Additional keywords arguments are passed to the output comm.

    Attributes:
        response_kwargs (dict): Keyword arguments for the response comm.
        request_order (list): Order of request IDs.
        responses (dict): Mapping between request IDs and response messages.
        ocomm (Comm): Request comm.
        icomm (Comm): Response comm.

    """

    _commtype = 'client'
    _dont_register = True
    
    def __init__(self, name, request_commtype=None, response_kwargs=None,
                 dont_open=False, is_async=False, direct_connection=False,
                 **kwargs):
        if response_kwargs is None:
            response_kwargs = dict()
        ocomm_name = name
        ocomm_kwargs = kwargs
        ocomm_kwargs['direction'] = 'send'
        ocomm_kwargs['dont_open'] = True
        ocomm_kwargs['commtype'] = request_commtype
        ocomm_kwargs.setdefault('use_async', is_async)
        if direct_connection:
            ocomm_kwargs.setdefault('is_client', True)
        self.direct_connection = direct_connection
        self.response_kwargs = response_kwargs
        self.ocomm = get_comm(ocomm_name, **ocomm_kwargs)
        self.icomm = None
        self.request_order = []
        self.responses = OrderedDict()
        self.response_kwargs.setdefault('is_interface', self.ocomm.is_interface)
        self.response_kwargs.setdefault('commtype', self.ocomm._commtype)
        self.response_kwargs.setdefault('recv_timeout', self.ocomm.recv_timeout)
        self.response_kwargs.setdefault('language', self.ocomm.language)
        self.response_kwargs.setdefault('use_async', self.ocomm.is_async)
        super(ClientComm, self).__init__(self.ocomm.name, dont_open=dont_open,
                                         recv_timeout=self.ocomm.recv_timeout,
                                         is_interface=self.ocomm.is_interface,
                                         direction='send', no_suffix=True,
                                         address=self.ocomm.address,
                                         is_async=self.ocomm.is_async)

    def get_status_message(self, nindent=0, **kwargs):
        r"""Return lines composing a status message.
        
        Args:
            nindent (int, optional): Number of tabs that should
                be used to indent each line. Defaults to 0.
            *kwargs: Additional arguments are passed to the
                parent class's method.
                
        Returns:
            tuple(list, prefix): Lines composing the status message and the
                prefix string used for the last message.

        """
        lines, prefix = super(ClientComm, self).get_status_message(
            nindent=nindent, **kwargs)
        lines.append('%s%-15s:' % (prefix, 'request comm'))
        lines += self.ocomm.get_status_message(nindent=(nindent + 1))[0]
        lines.append('%s%-15s:' % (prefix, 'response comms'))
        lines += self.icomm.get_status_message(nindent=(nindent + 1))[0]
        return lines, prefix
    
    @classmethod
    def is_installed(cls, language=None):
        r"""Determine if the necessary libraries are installed for this
        communication class.

        Args:
            language (str, optional): Specific language that should be checked
                for compatibility. Defaults to None and all languages supported
                on the current platform will be checked.

        Returns:
            bool: Is the comm installed.

        """
        return import_comm().is_installed(language=language)

    @property
    def maxMsgSize(self):
        r"""int: Maximum size of a single message that should be sent."""
        return self.ocomm.maxMsgSize
        
    @classmethod
    def underlying_comm_class(self):
        r"""str: Name of underlying communication class."""
        return import_comm().underlying_comm_class()

    @classmethod
    def comm_count(cls):
        r"""int: Number of communication connections."""
        return import_comm().comm_count()

    @classmethod
    def new_comm_kwargs(cls, name, request_commtype=None, **kwargs):
        r"""Initialize communication with new comms.

        Args:
            name (str): Name for new comm.
            request_commtype (str, optional): Name of class for new output
                comm. Defaults to None.

        """
        args = [name]
        ocomm_class = import_comm(request_commtype)
        kwargs['direction'] = 'send'
        if 'address' not in kwargs:
            oargs, kwargs = ocomm_class.new_comm_kwargs(name, **kwargs)
        kwargs['request_commtype'] = request_commtype
        return args, kwargs

    @property
    def opp_address(self):
        r"""str: Address for opposite comm."""
        return self.ocomm.opp_address
        
    @property
    def opp_comms(self):
        r"""dict: Name/address pairs for opposite comms."""
        out = super(ClientComm, self).opp_comms
        out.update(**self.ocomm.opp_comms)
        return out

    def opp_comm_kwargs(self):
        r"""Get keyword arguments to initialize communication with opposite
        comm object.

        Returns:
            dict: Keyword arguments for opposite comm object.

        """
        kwargs = super(ClientComm, self).opp_comm_kwargs()
        kwargs['commtype'] = "server"
        kwargs['request_commtype'] = self.ocomm._commtype
        kwargs['response_kwargs'] = self.response_kwargs
        kwargs['direct_connection'] = self.direct_connection
        return kwargs
        
    def open(self):
        r"""Open the connection."""
        super(ClientComm, self).open()
        self.ocomm.open()

    def close(self, *args, **kwargs):
        r"""Close the connection."""
        self.ocomm.close(*args, **kwargs)
        if self.icomm is not None:
            self.icomm.close(*args, **kwargs)
        super(ClientComm, self).close(*args, **kwargs)

    @property
    def is_open(self):
        r"""bool: True if the connection is open."""
        return self.ocomm.is_open

    @property
    def is_closed(self):
        r"""bool: True if the connection is closed."""
        return self.ocomm.is_closed

    @property
    def n_msg_send(self):
        r"""int: The number of messages in the connection."""
        return self.ocomm.n_msg_send

    @property
    def n_msg_send_drain(self):
        r"""int: The number of outgoing messages in the connection to drain."""
        return self.ocomm.n_msg_send_drain

    def atexit(self):  # pragma: debug
        r"""Close operations."""
        self.send_eof()
        super(ClientComm, self).atexit()
        
    # RESPONSE COMM
    def create_response_comm(self):
        r"""Create a response comm based on information from the last header."""
        header = dict(request_id=str(uuid.uuid4()))
        while header['request_id'] in self.request_order:  # pragma: debug
            header['request_id'] += str(uuid.uuid4())
        if self.icomm is None:
            comm_kwargs = dict(direction='recv', is_response_client=True,
                               **self.response_kwargs)
            if comm_kwargs.get('use_async', False):
                comm_kwargs['async_recv_method'] = 'recv_message'
            self.icomm = new_comm(self.name + '.client_response_comm', **comm_kwargs)
        header['response_address'] = self.icomm.address
        self.request_order.append(header['request_id'])
        return header

    # SEND METHODS
    def prepare_message(self, *args, **kwargs):
        r"""Perform actions preparing to send a message.

        Args:
            *args: Components of the outgoing message.
            **kwargs: Keyword arguments are passed to the request comm's
                prepare_message method.

        Returns:
            CommMessage: Serialized and annotated message.

        """
        def add_response_address(msg):
            if msg.flag == CommBase.FLAG_SUCCESS:
                if msg.header is None:
                    msg.header = {}
                msg.header.update(self.create_response_comm())
            return msg
        kwargs.setdefault('after_prepare_message', [])
        kwargs['after_prepare_message'].append(add_response_address)
        return self.ocomm.prepare_message(*args, **kwargs)
        
    def send_message(self, msg, **kwargs):
        r"""Send a message encapsulated in a CommMessage object.

        Args:
            msg (CommMessage): Message to be sent.
            **kwargs: Additional keyword arguments are passed to the request
                comm's send_message method.

        Returns:
            bool: Success or failure of send.
        
        """
        out = self.ocomm.send_message(msg, **kwargs)
        self.errors += self.ocomm.errors
        return out
        
    # RECV METHODS
    def recv_message(self, *args, **kwargs):
        r"""Receive a message.

        Args:
            *args: Arguments are passed to the response comm's recv_message method.
            **kwargs: Keyword arguments are passed to the response comm's recv_message
                method.

        Returns:
            CommMessage: Received message.

        """
        if not self.request_order:  # pragma: debug
            raise RuntimeError("There are not any requests registered.")
        while self.responses.get(self.request_order[0], None) is None:
            msg = self.icomm.recv_message(*args, **kwargs)
            self.errors += self.icomm.errors
            if msg.flag != CommBase.FLAG_SUCCESS:  # pragma: debug
                break
            assert(msg.header['request_id'] not in self.responses)
            self.responses[msg.header['request_id']] = msg
        if self.request_order[0] in self.responses:
            msg = self.responses.pop(self.request_order[0])
            self.request_order.pop(0)
        return msg

    def finalize_message(self, msg, **kwargs):
        r"""Perform actions to decipher a message.

        Args:
            msg (CommMessage): Initial message object to be finalized.
            **kwargs: Keyword arguments are passed to the response comm's
                finalize_message method.

        Returns:
            CommMessage: Deserialized and annotated message.

        """
        if self.icomm is None:  # pragma: debug
            raise RuntimeError("There are not any registered response comms.")
        return self.icomm.finalize_message(msg, **kwargs)
        
    # CALL
    def call(self, *args, **kwargs):
        r"""Do RPC call. The request message is sent to the output comm and the
        response is received from the input comm.

        Args:
            *args: Arguments are passed to output comm send method.
            **kwargs: Keyword arguments are passed to output comm send method

        Returns:
            obj: Output from input comm recv method.

        """
        flag = self.send(*args, **kwargs)
        if not flag:  # pragma: debug
            return (False, self.empty_obj_recv)
        return self.recv(timeout=False)

    def call_nolimit(self, *args, **kwargs):
        r"""Alias for call."""
        return self.call(*args, **kwargs)

    # OLD STYLE ALIASES
    def rpcSend(self, *args, **kwargs):
        r"""Alias for RPCComm.send"""
        return self.send(*args, **kwargs)

    def rpcRecv(self, *args, **kwargs):
        r"""Alias for RPCComm.recv"""
        return self.recv(*args, **kwargs)
    
    def rpcCall(self, *args, **kwargs):
        r"""Alias for RPCComm.call"""
        return self.call(*args, **kwargs)
    
    def drain_messages(self, direction='send', **kwargs):
        r"""Sleep while waiting for messages to be drained."""
        if direction == 'send':
            self.ocomm.drain_messages(direction='send', **kwargs)



from yggdrasil.drivers.InterpretedModelDriver import InterpretedModelDriver


class FunctionModelDriver(InterpretedModelDriver):
    r"""Class that acts as a functional interface to input/output channels
    that are not connected.

    Args:

    Attributes:

    """
    executable_type = 'other'
    language = 'function'
    full_language = False
    base_languages = ['python']
    language_ext = []
    no_executable = True
    comms_implicit = True

    @classmethod
    def is_language_installed(self):
        r"""Determine if this model driver is installed on the current
        machine.

        Returns:
            bool: Truth of if this model driver can be run on the current
                machine.

        """
        # This is being run so python exists
        return True

    @classmethod
    def is_configured(cls):
        r"""Determine if the appropriate configuration has been performed (e.g.
        installation of supporting libraries etc.)

        Returns:
            bool: True if the language has been configured.

        """
        # There are not any config options
        return True

    @classmethod
    def language_version(cls, **kwargs):
        r"""Determine the version of this language.

        Args:
            **kwargs: Keyword arguments are passed to cls.run_executable.

        Returns:
            str: Version of compiler/interpreter for this language.

        """
        return '0'
    
    def before_start(self):
        r"""Actions to perform before the run starts."""
        pass
                                
    def before_loop(self):
        r"""Actions before loop."""
        pass

    def run_loop(self):
        r"""Loop to check if model is still running and forward output."""
        pass



import os
import uuid
from collections import OrderedDict
from yggdrasil.communication import CommBase, get_comm, import_comm


class Request(object):

    __slots__ = ['response_address', 'request_id']

    def __init__(self, response_address, request_id):
        self.response_address = response_address
        self.request_id = request_id
        super(Request, self).__init__()


class ServerComm(CommBase.CommBase):
    r"""Class for handling Server side communication.

    Args:
        name (str): The environment variable where communication address is
            stored.
        request_commtype (str, optional): Comm class that should be used for
            the request comm. Defaults to None.
        response_kwargs (dict, optional): Keyword arguments for the response
            comm. Defaults to empty dict.
        direct_connection (bool, optional): If True, the comm will be
            directly connected to a ServerComm. Defaults to False.
        **kwargs: Additional keywords arguments are passed to the input comm.

    Attributes:
        response_kwargs (dict): Keyword arguments for the response comm.
        icomm (Comm): Request comm.
        ocomm (OrderedDict): Response comms for each request.

    """

    _commtype = 'server'
    _dont_register = True
    
    def __init__(self, name, request_commtype=None, response_kwargs=None,
                 dont_open=False, is_async=False, direct_connection=False,
                 **kwargs):
        if response_kwargs is None:
            response_kwargs = dict()
        icomm_name = name
        icomm_kwargs = kwargs
        icomm_kwargs.update(direction='recv',
                            dont_open=True,
                            commtype=request_commtype)
        icomm_kwargs.setdefault('is_server', True)
        icomm_kwargs.setdefault('use_async', is_async)
        if icomm_kwargs.get('use_async', False):
            icomm_kwargs.setdefault('async_recv_method', 'recv_message')
        self.direct_connection = direct_connection
        self.response_kwargs = response_kwargs
        self.icomm = get_comm(icomm_name, **icomm_kwargs)
        self.ocomm = OrderedDict()
        self.requests = OrderedDict()
        self.response_kwargs.setdefault('is_interface', self.icomm.is_interface)
        self.response_kwargs.setdefault('commtype', self.icomm._commtype)
        self.response_kwargs.setdefault('recv_timeout', self.icomm.recv_timeout)
        self.response_kwargs.setdefault('language', self.icomm.language)
        self.response_kwargs.setdefault('use_async', self.icomm.is_async)
        self.clients = []
        self.closed_clients = []
        self.nclients_expected = int(os.environ.get('YGG_NCLIENTS', 0))
        super(ServerComm, self).__init__(self.icomm.name, dont_open=dont_open,
                                         recv_timeout=self.icomm.recv_timeout,
                                         is_interface=self.icomm.is_interface,
                                         direction='recv', no_suffix=True,
                                         address=self.icomm.address,
                                         is_async=self.icomm.is_async)

    def get_status_message(self, nindent=0, **kwargs):
        r"""Return lines composing a status message.
        
        Args:
            nindent (int, optional): Number of tabs that should
                be used to indent each line. Defaults to 0.
            *kwargs: Additional arguments are passed to the
                parent class's method.
                
        Returns:
            tuple(list, prefix): Lines composing the status message and the
                prefix string used for the last message.

        """
        lines, prefix = super(ServerComm, self).get_status_message(
            nindent=nindent, **kwargs)
        lines.append('%s%-15s:' % (prefix, 'request comm'))
        lines += self.icomm.get_status_message(nindent=(nindent + 1))[0]
        lines.append('%s%-15s:' % (prefix, 'response comms'))
        for x in self.ocomm.values():
            lines += x.get_status_message(nindent=(nindent + 1))[0]
        return lines, prefix
        
    @classmethod
    def is_installed(cls, language=None):
        r"""Determine if the necessary libraries are installed for this
        communication class.

        Args:
            language (str, optional): Specific language that should be checked
                for compatibility. Defaults to None and all languages supported
                on the current platform will be checked.

        Returns:
            bool: Is the comm installed.

        """
        return import_comm().is_installed(language=language)

    @property
    def maxMsgSize(self):
        r"""int: Maximum size of a single message that should be sent."""
        return self.icomm.maxMsgSize
        
    @classmethod
    def underlying_comm_class(self):
        r"""str: Name of underlying communication class."""
        return import_comm().underlying_comm_class()

    @classmethod
    def comm_count(cls):
        r"""int: Number of communication connections."""
        return import_comm().comm_count()

    @classmethod
    def new_comm_kwargs(cls, name, request_commtype=None, **kwargs):
        r"""Initialize communication with new comms.

        Args:
            name (str): Name for new comm.
            request_commtype (str, optional): Name of class for new input
                comm. Defaults to None.

        """
        args = [name]
        icomm_class = import_comm(request_commtype)
        kwargs['direction'] = 'recv'
        if 'address' not in kwargs:
            iargs, kwargs = icomm_class.new_comm_kwargs(name, **kwargs)
        kwargs['request_commtype'] = request_commtype
        return args, kwargs

    @property
    def opp_comms(self):
        r"""dict: Name/address pairs for opposite comms."""
        out = super(ServerComm, self).opp_comms
        out.update(**self.icomm.opp_comms)
        return out

    def opp_comm_kwargs(self):
        r"""Get keyword arguments to initialize communication with opposite
        comm object.

        Returns:
            dict: Keyword arguments for opposite comm object.

        """
        kwargs = super(ServerComm, self).opp_comm_kwargs()
        kwargs['commtype'] = "client"
        kwargs['request_commtype'] = self.icomm._commtype
        kwargs['response_kwargs'] = self.response_kwargs
        kwargs['direct_connection'] = self.direct_connection
        return kwargs
        
    def open(self):
        r"""Open the connection."""
        super(ServerComm, self).open()
        self.icomm.open()

    def close(self, *args, **kwargs):
        r"""Close the connection."""
        self.icomm.close(*args, **kwargs)
        for ocomm in self.ocomm.values():
            ocomm.close()
        super(ServerComm, self).close(*args, **kwargs)

    @property
    def is_open(self):
        r"""bool: True if the connection is open."""
        return self.icomm.is_open

    @property
    def is_closed(self):
        r"""bool: True if the connection is closed."""
        return self.icomm.is_closed

    @property
    def n_msg_recv(self):
        r"""int: The number of messages in the connection."""
        return self.icomm.n_msg_recv

    @property
    def n_msg_recv_drain(self):
        r"""int: The number of messages in the connection to drain."""
        return self.icomm.n_msg_recv_drain

    @property
    def open_clients(self):
        r"""list: Available open clients."""
        return list(set(self.clients) - set(self.closed_clients))

    @property
    def all_clients_connected(self):
        r"""bool: True if all expected clients have connected.
        False otherwise."""
        return ((self.nclients_expected > 0)
                and (len(self.clients) >= self.nclients_expected))

    # RESPONSE COMM
    def create_response_comm(self, header):
        r"""Create a response comm based on information from the last header."""
        if not isinstance(header, dict):  # pragma: debug
            raise RuntimeError("No header received with last message.")
        elif 'response_address' not in header:  # pragma: debug
            raise RuntimeError("Last header does not contain response address.")
        comm_kwargs = dict(address=header['response_address'],
                           direction='send',
                           **self.response_kwargs)
        if self.direct_connection:
            comm_kwargs['is_response_client'] = True
        else:
            comm_kwargs['is_response_server'] = True
        response_id = header['request_id']
        while response_id in self.requests:  # pragma: debug
            response_id += str(uuid.uuid4())
        header['response_id'] = response_id
        if header['response_address'] not in self.ocomm:
            self.ocomm[header['response_address']] = get_comm(
                self.name + '.server_response_comm.' + response_id,
                **comm_kwargs)
            client_model = header.get('model', '')
            self.ocomm[header['response_address']].client_model = client_model
            if client_model and (client_model not in self.clients):
                self.clients.append(client_model)
        self.requests[response_id] = Request(header['response_address'],
                                             header['request_id'])

    # SEND METHODS
    def send_to(self, response_id, *args, **kwargs):
        r"""Send a message to a specific response comm.

        Args:
            response_id (str): ID used to register the response comm.
            *args: Arguments are passed to output comm send method.
            **kwargs: Keyword arguments are passed to output comm send method.
    
        Returns:
            obj: Output from output comm send method.

        """
        kwargs.setdefault('header_kwargs', {})
        kwargs['header_kwargs']['response_id'] = response_id
        return self.send(*args, **kwargs)
        
    def prepare_message(self, *args, **kwargs):
        r"""Perform actions preparing to send a message.

        Args:
            *args: Components of the outgoing message.
            **kwargs: Keyword arguments are passed to the request comm's
                prepare_message method.

        Returns:
            CommMessage: Serialized and annotated message.

        """
        if len(self.requests) == 0:  # pragma: debug
            raise RuntimeError("There is no registered request.")
        kwargs.setdefault('header_kwargs', {})
        response_id = kwargs['header_kwargs'].get('response_id', None)
        if response_id is None:
            response_id = next(iter(self.requests.keys()))
            kwargs['header_kwargs']['response_id'] = response_id
        request = self.requests[response_id]
        kwargs['header_kwargs']['request_id'] = request.request_id
        return self.ocomm[request.response_address].prepare_message(
            *args, **kwargs)
        
    def send_message(self, msg, **kwargs):
        r"""Send a message encapsulated in a CommMessage object.

        Args:
            msg (CommMessage): Message to be sent.
            **kwargs: Additional keyword arguments are passed to the response
                comm's send_message method.

        Returns:
            bool: Success or failure of send.
        
        """
        response_id = msg.header['response_id']
        response_comm = self.ocomm[self.requests[response_id].response_address]
        out = response_comm.send_message(msg, **kwargs)
        self.errors += response_comm.errors
        if out:
            self.requests.pop(response_id)
        return out
                                                  
    # RECV METHODS
    def recv_from(self, *args, **kwargs):
        r"""Receive a message from the input comm and open a new response comm
        for output using address from the header, returning the response_id.

        Args:
            *args: Arguments are passed to input comm recv method.
            **kwargs: Keyword arguments are passed to input comm recv method.

        Returns:
            tuple(bool, obj, str): Success or failure of recv call,
                output from input comm recv method, and response_id that
                response should be sent to.

        """
        return_message_object = kwargs.pop('return_message_object', False)
        kwargs['return_message_object'] = True
        response_id = None
        out = self.recv(*args, **kwargs)
        if not return_message_object:
            if out.flag == CommBase.FLAG_SUCCESS:
                response_id = out.header['response_id']
            out = (bool(out.flag), out.args, response_id)
        return out
    
    def recv_message(self, *args, **kwargs):
        r"""Receive a message.

        Args:
            *args: Arguments are passed to the request comm's recv_message method.
            **kwargs: Keyword arguments are passed to the request comm's recv_message
                method.

        Returns:
            CommMessage: Received message.

        """
        out = self.icomm.recv_message(*args, **kwargs)
        self.errors += self.icomm.errors
        return out
        
    def finalize_message(self, msg, **kwargs):
        r"""Perform actions to decipher a message.

        Args:
            msg (CommMessage): Initial message object to be finalized.
            **kwargs: Keyword arguments are passed to the request comm's
                finalize_message method.

        Returns:
            CommMessage: Deserialized and annotated message.

        """
        def check_for_client_info(msg):
            if msg.flag == CommBase.FLAG_SUCCESS:
                if isinstance(msg.args, bytes) and (msg.args == CommBase.YGG_CLIENT_EOF):
                    self.debug("Client signed off: %s", msg.header['model'])
                    self.closed_clients.append(msg.header['model'])
                    msg.flag = CommBase.FLAG_SKIP
                else:
                    self.create_response_comm(msg.header)
            return msg
        out = self.icomm.finalize_message(msg, **kwargs)
        return check_for_client_info(out)
        
    # OLD STYLE ALIASES
    def rpcSend(self, *args, **kwargs):
        r"""Alias for RPCComm.send"""
        return self.send(*args, **kwargs)

    def rpcRecv(self, *args, **kwargs):
        r"""Alias for RPCComm.recv"""
        return self.recv(*args, **kwargs)
        
    def drain_messages(self, direction='recv', **kwargs):
        r"""Sleep while waiting for messages to be drained."""
        if direction == 'recv':
            self.icomm.drain_messages(direction='recv', **kwargs)
            self.errors += self.icomm.errors

    def purge(self):
        r"""Purge input and output comms."""
        self.icomm.purge()
        # Not sure if server should purge the response queue...
        # for ocomm in self.ocomm.values():
        #     ocomm.purge()
        super(ServerComm, self).purge()
    
    def drain_server_signon_messages(self, **kwargs):
        r"""Drain server signon messages. This should only be used
        for testing purposes."""
        self.icomm.drain_server_signon_messages(**kwargs)



from yggdrasil.drivers.ConnectionDriver import ConnectionDriver, run_remotely
from yggdrasil.drivers.RPCResponseDriver import RPCResponseDriver
from yggdrasil.communication import CommBase


class RPCRequestDriver(ConnectionDriver):
    r"""Class for handling client side RPC type communication.

    Args:
        model_request_name (str): The name of the channel used by the client
            model to send requests.
        **kwargs: Additional keyword arguments are passed to parent class.

    Attributes:
        response_drivers (dict): Response drivers created for each request.

    """

    _connection_type = 'rpc_request'

    def __init__(self, model_request_name, response_kwargs=None, **kwargs):
        # Input communicator
        inputs = kwargs.get('inputs', [{}])
        # inputs[0]['name'] = model_request_name + '.client_model_request'
        kwargs['inputs'] = inputs
        # Output communicator
        outputs = kwargs.get('outputs', [{}])
        # outputs[0]['name'] = model_request_name + '.server_model_request'
        outputs[0]['is_client'] = True
        outputs[0]['close_on_eof_send'] = False
        kwargs['outputs'] = outputs
        if response_kwargs is None:
            response_kwargs = {}
        self.response_kwargs = response_kwargs
        # Parent and attributes
        super(RPCRequestDriver, self).__init__(model_request_name, **kwargs)
        self.response_kwargs.setdefault('commtype', self.ocomm._commtype)
        self.response_drivers = {}
        self._block_response = False

    @property
    def servers_recvd(self):
        r"""list: Names of server models that have returned responses."""
        out = {}
        for x in self.response_drivers.values():
            for k, v in x.models_recvd.items():
                out.setdefault(k, 0)
                out[k] += v
        return out

    @property
    @run_remotely
    def clients(self):
        r"""list: Clients that are connected."""
        return self.models['input'].copy()

    @property
    @run_remotely
    def nclients(self):
        r"""int: Number of clients that are connected."""
        return len(self.clients)

    @property
    def model_env(self):
        r"""dict: Mapping between model name and opposite comm
        environment variables that need to be provided to the model."""
        out = super(RPCRequestDriver, self).model_env
        # Add is_rpc flag to output model env variables
        for k in self.ocomm.model_env.keys():
            out[k]['YGG_IS_SERVER'] = 'True'
        return out
        
    def close_response_drivers(self):
        r"""Close response driver."""
        with self.lock:
            self.debug("Closing response drivers.")
            self._block_response = True
            for x in self.response_drivers.values():
                x.terminate()
            self.response_drivers = {}

    def close_comm(self):
        r"""Close response drivers."""
        self.close_response_drivers()
        super(RPCRequestDriver, self).close_comm()
            
    def printStatus(self, *args, **kwargs):
        r"""Also print response drivers."""
        super(RPCRequestDriver, self).printStatus(*args, **kwargs)
        for x in self.response_drivers.values():
            x.printStatus(*args, **kwargs)

    @run_remotely
    def remove_model(self, direction, name):
        r"""Remove a model from the list of models.

        Args:
            direction (str): Direction of model.
            name (str): Name of model exiting.

        Returns:
            bool: True if all of the input/output models have signed
                off; False otherwise.

        """
        with self.lock:
            clients = self.clients
            if (direction == "input") and (name in clients) and (len(clients) > 1):
                super(RPCRequestDriver, self).send_message(
                    CommBase.CommMessage(args=CommBase.YGG_CLIENT_EOF,
                                         flag=CommBase.FLAG_SUCCESS),
                    header_kwargs={'raw': True, 'model': name},
                    skip_processing=True)
            out = super(RPCRequestDriver, self).remove_model(
                direction, name)
            if out:
                self.send_eof()
            return out
        
    # def send_eof(self):
    #     r"""Send EOF message.

    #     Returns:
    #         bool: Success or failure of send.

    #     """
    #     if self.ocomm.partner_copies > 1:
    #         self.ocomm.partner_copies = len(self.servers_recvd)
    #     return super(RPCRequestDriver, self).send_eof()
        
    def on_eof(self, msg):
        r"""On EOF, decrement number of clients. Only send EOF if the number
        of clients drops to 0.

        Args:
            msg (CommMessage): Message object that provided the EOF.

        Returns:
            CommMessage, bool: Value that should be returned by recv_message on EOF.

        """
        with self.lock:
            self.remove_model('input', msg.header.get('model', ''))
            if self.nclients == 0:
                self.debug("All clients have signed off (EOF).")
                return super(RPCRequestDriver, self).on_eof(msg)
        return CommBase.CommMessage(flag=CommBase.FLAG_EMPTY,
                                    args=self.icomm.empty_obj_recv)

    def before_loop(self):
        r"""Send client sign on to server response driver."""
        super(RPCRequestDriver, self).before_loop()
        self.ocomm._send_serializer = True

    def send_message(self, msg, **kwargs):
        r"""Start a response driver for a request message and send message with
        header.

        Args:
            msg (CommMessage): Message being sent.
            **kwargs: Keyword arguments are passed to parent class send_message.

        Returns:
            bool: Success or failure of send.

        """
        if self.ocomm.is_closed:
            return False
        # Start response driver
        if msg.flag != CommBase.FLAG_EOF:
            with self.lock:
                if (not self.is_comm_open) or self._block_response:  # pragma: debug
                    self.debug("Comm closed, not creating response driver.")
                    return False
                
                if msg.header['response_address'] in self.response_drivers:
                    response_driver = self.response_drivers[
                        msg.header['response_address']]
                else:
                    drv_args = [msg.header['response_address'],
                                msg.header['request_id']]
                    drv_kwargs = dict(
                        request_name=self.name,
                        inputs=[self.response_kwargs.copy()],
                        outputs=[{'commtype': msg.header["commtype"]}])
                    self.debug("Creating response comm: address = %s, request_id = %s",
                               msg.header['response_address'],
                               msg.header['request_id'])
                    try:
                        response_driver = RPCResponseDriver(
                            *drv_args, **drv_kwargs)
                        self.response_drivers[
                            msg.header['response_address']] = response_driver
                        response_driver.start()
                        self.debug("Started response comm: address = %s, request_id = %s",
                                   msg.header['response_address'],
                                   msg.header['request_id'])
                    except BaseException:  # pragma: debug
                        self.exception("Could not create/start response driver.")
                        return False
            # Send response address in header
            kwargs.setdefault('header_kwargs', {})
            kwargs['header_kwargs'].setdefault(
                'response_address', response_driver.response_address)
            kwargs['header_kwargs'].setdefault('request_id', msg.header['request_id'])
            kwargs['header_kwargs'].setdefault('model', msg.header.get('model', ''))
        return super(RPCRequestDriver, self).send_message(msg, **kwargs)

    def run_loop(self):
        r"""Run the driver. Continue looping over messages until there are not
        any left or the communication channel is closed.
        """
        super(RPCRequestDriver, self).run_loop()
        if not self.was_break:
            self.prune_response_drivers()

    def prune_response_drivers(self):
        r"""Remove response drivers that are no longer being used."""
        with self.lock:
            # remove_idx = []
            # for i, x in enumerate(self.response_drivers):
            for x in self.response_drivers.values():
                self.errors += x.errors
            #     if (((not x.is_alive())
            #          and x.icomm.is_confirmed_recv
            #          and x.ocomm.is_confirmed_send)):
            #         x.cleanup()
            #         remove_idx.append(i)
            # for i in remove_idx[::-1]:
            #     self.response_drivers.pop(i)



import copy
from yggdrasil.communication.tests import test_CommBase as parent


class TestValueComm(parent.TestCommBase):
    r"""Test for ValueComm communication class."""

    comm = 'ValueComm'
    attr_list = (copy.deepcopy(parent.TestCommBase.attr_list)
                 + ['remaining'])

    # Disable tests that don't make sense for ValueComm
    test_error_send = None
    test_error_recv = None
    test_work_comm = None
    test_drain_messages = None
    test_recv_nomsg = None
    test_send_recv_filter_eof = None
    test_send_recv_filter_pass = None
    test_send_recv_filter_send_filter = None
    test_send_recv_filter_recv_filter = None
    test_send_recv_nolimit = None
    test_send_recv_array = None
    test_eof = None
    test_eof_no_close = None
    test_send_recv_dict = None
    test_send_recv_dict_names = None

    @property
    def send_inst_kwargs(self):
        r"""dict: Keyword arguments for send instance."""
        out = super(TestValueComm, self).send_inst_kwargs
        out['direction'] = 'recv'
        return out
    
    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for tested class."""
        return self.send_inst_kwargs

    def test_send_recv(self):
        r"""Test send/recv of a small message."""
        n_recv = self.testing_options['kwargs']['count']
        msg_recv = self.test_msg
        if self.use_async:
            assert(self.recv_instance.n_msg_recv > 0)
        else:
            self.assert_equal(self.recv_instance.n_msg_recv, n_recv)
        # Wait for messages to be received
        for i in range(n_recv):
            flag, msg_recv0 = self.recv_instance.recv(self.timeout)
            assert(flag)
            self.assert_msg_equal(msg_recv0, msg_recv)
        # Receive after empty
        assert(self.recv_instance.is_open)
        flag, msg_recv0 = self.recv_instance.recv(self.timeout)
        assert(not flag)
        assert(self.recv_instance.is_eof(msg_recv0))
        # assert(flag)
        # assert(self.recv_instance.is_empty_recv(msg_recv0))
        # Confirm recept of messages
        self.recv_instance.wait_for_confirm(timeout=self.timeout)
        assert(self.recv_instance.is_confirmed)
        self.recv_instance.confirm(noblock=True)
        self.assert_equal(self.recv_instance.n_msg_recv, 0)

    def test_send_recv_after_close(self):
        r"""Test that opening twice dosn't cause errors and that send/recv after
        close returns false."""
        self.recv_instance.open()
        self.recv_instance.close()
        assert(self.recv_instance.is_closed)
        flag, msg_recv = self.recv_instance.recv()
        if not self.use_async:
            assert(not flag)
        self.assert_raises(RuntimeError, self.recv_instance.send, None)

    def test_purge(self, nrecv=1):
        r"""Test purging messages from the comm."""
        assert(self.recv_instance.n_msg > 0)
        self.recv_instance.purge()
        self.assert_equal(self.recv_instance.n_msg, 0)
        self.recv_instance.close()
        self.recv_instance.purge()



import copy
import numpy as np
from yggdrasil.tests import assert_raises
from yggdrasil.serialize import AsciiTableSerialize
from yggdrasil.serialize.tests import test_DefaultSerialize as parent


def test_serialize_nofmt():
    r"""Test error on serialization without a format."""
    inst = AsciiTableSerialize.AsciiTableSerialize()
    inst.initialized = True
    test_msg = np.zeros((5, 5))
    assert_raises(RuntimeError, inst.serialize, test_msg)

    
def test_deserialize_nofmt():
    r"""Test error on deserialization without a format."""
    inst = AsciiTableSerialize.AsciiTableSerialize()
    test_msg = b'lskdbjs;kfbj'
    test_msg = inst.encoded_datatype.serialize(test_msg, metadata={})
    assert_raises(RuntimeError, inst.deserialize, test_msg)


class TestAsciiTableSerialize(parent.TestDefaultSerialize):
    r"""Test class for AsciiTableSerialize class."""

    _cls = 'AsciiTableSerialize'
    attr_list = (copy.deepcopy(parent.TestDefaultSerialize.attr_list),
                 ['format_str', 'field_names', 'field_units', 'as_array'])
    
    def test_field_specs(self):
        r"""Test field specifiers."""
        if not self.instance.initialized:
            self.instance.serialize(self.testing_options['objects'][0],
                                    no_metadata=True)
        super(TestAsciiTableSerialize, self).test_field_specs()
        # Specific to this class
        if 'format_str' in self.testing_options:
            self.assert_equal(self.instance.format_str,
                              self.testing_options['format_str'].encode("utf-8"))
        field_names = self.testing_options.get('field_names', None)
        self.assert_equal(self.instance.field_names, field_names)
        field_units = self.testing_options.get('field_units', None)
        self.assert_equal(self.instance.field_units, field_units)


class TestAsciiTableSerializeSingle(parent.TestDefaultSerialize):
    r"""Test class for AsciiTableSerialize class."""

    _cls = 'AsciiTableSerialize'
    _empty_obj = []
    _objects = [(1, )]

    def __init__(self, *args, **kwargs):
        super(TestAsciiTableSerializeSingle, self).__init__(*args, **kwargs)
        self._inst_kwargs['format_str'] = '%d\n'

    def get_options(self):
        r"""Get testing options."""
        out = {'kwargs': {'format_str': '%d\n'},
               'empty': [],
               'objects': [(1, )],
               'extra_kwargs': {},
               'typedef': {'type': 'array',
                           'items': [{'type': 'int', 'precision': 32}]},
               'dtype': None,
               'is_user_defined': False}
        return out


class TestAsciiTableSerialize_asarray(TestAsciiTableSerialize):
    r"""Test class for AsciiTableSerialize class with as_array."""

    testing_option_kws = {'array_columns': True}


class TestAsciiTableSerialize_object(TestAsciiTableSerialize):
    r"""Test class for AsciiTableSerialize class with object."""

    testing_option_kws = {}

    def get_options(self):
        r"""Get testing options."""
        out = super(TestAsciiTableSerialize_object, self).get_options()
        out['kwargs'] = {'field_units': out['kwargs']['field_units']}
        out['format_str'] = '%s\t%d\t%g\n'
        out['field_names'] = ['%s_%s' % (k, x) for k, x
                              in zip('abc', out['field_names'])]
        out['objects'] = [{k: ix for k, ix in zip(out['field_names'], x)}
                          for x in out['objects']]
        for x, k2 in zip(out['typedef']['items'], out['field_names']):
            out['contents'].replace(x['title'].encode("utf-8"),
                                    k2.encode("utf-8"))
            x['title'] = k2
        return out

    def map_sent2recv(self, obj):
        r"""Convert a sent object into a received one."""
        out = super(TestAsciiTableSerialize_object, self).map_sent2recv(obj)
        return self.instance.datatype.coerce_type(
            out, typedef=self.instance.typedef)



import subprocess
import os
from contextlib import contextmanager
from yggdrasil.components import import_component


_temp_error_registry = {}


class TemporaryCommunicationError(Exception):
    r"""Raised when the comm is open, but send/recv is temporarily disabled.

    Args:
        msg (str): Error message.
        max_consecutive_allowed (int, optional): Maximum number of times this
            error should be raised as a TemporaryCommunicationError before
            being elevated to a FatalCommunicationError. Defaults to None and
            is never elevated.
        registry_key (str, optional): Key that should be used to register the
            error. Defaults to msg.

    Raises:
        FatalCommunicationError: If the error is raised more than the number
            of times specified via max_consecutive_allowed.

    """

    def __init__(self, msg, max_consecutive_allowed=None,
                 registry_key=None, **kwargs):
        super(TemporaryCommunicationError, self).__init__(msg, **kwargs)
        self.max_consecutive_allowed = max_consecutive_allowed
        if max_consecutive_allowed is not None:
            global _temp_error_registry
            assert(registry_key is not None)
            _temp_error_registry.setdefault(registry_key, 0)
            _temp_error_registry[registry_key] += 1
            if ((_temp_error_registry[registry_key]
                 > max_consecutive_allowed)):  # pragma: debug
                raise FatalCommunicationError(msg, **kwargs)

    @classmethod
    def reset(cls, registry_key):
        r"""Reset the registry for a TemporaryCommunicationError."""
        global _temp_error_registry
        _temp_error_registry.pop(registry_key, None)


class NoMessages(TemporaryCommunicationError):
    r"""Raised when the comm is open, but there are no messages waiting."""
    pass


class FatalCommunicationError(Exception):
    r"""Raised when the comm cannot recover."""
    pass


def import_comm(commtype=None):
    r"""Import a comm class from its component subtype.

    Args:
        commtype (str, optional): Communication class subtype. Defaults to
            the default comm type for the current OS.

    Returns:
        CommBase: Associated communication class.

    """
    if commtype in ['server', 'client', 'fork']:
        commtype = '%sComm' % commtype.title()
    return import_component('comm', commtype)


def determine_suffix(no_suffix=False, reverse_names=False,
                     direction='send', **kwargs):
    r"""Determine the suffix that should be used for the comm name.

    Args:
        no_suffix (bool, optional): If True, the suffix will be an empty
            string. Defaults to False.
        reverse_names (bool, optional): If True, the suffix will be
            opposite that indicated by the direction. Defaults to False.
        direction (str, optional): The direction that the comm will
            processing messages. Defaults to 'send'.
        **kwargs: Additional keyword arguments are ignored.

    Returns:
        str: Suffix that will be added to the comm name when producing
            the name of the environment variable where information about
            the comm will be stored.

    Raises:
        ValueError: If the direction is not 'recv' or 'send'.

    """
    if direction not in ['send', 'recv']:
        raise ValueError("Unrecognized message direction: %s" % direction)
    if no_suffix:
        suffix = ''
    else:
        if ((((direction == 'send') and (not reverse_names))
             or ((direction == 'recv') and reverse_names))):
            suffix = '_OUT'
        else:
            suffix = '_IN'
    return suffix


def new_comm(name, commtype=None, use_async=False, **kwargs):
    r"""Return a new communicator, creating necessary components for
    communication (queues, sockets, channels, etc.).

    Args:
        name (str): Communicator name.
        commtype (str, list, optional): Name of communicator type or a list
            of specifiers for multiple communicators that should be joined.
            Defaults to None.
        use_async (bool, optional): If True, send/recv operations will
            be performed asynchronously on new threads. Defaults to
            False.
        **kwargs: Additional keyword arguments are passed to communicator
            class method new_comm.

    Returns:
        Comm: Communicator of given class.

    """
    assert('comm' not in kwargs)
    if isinstance(commtype, list):
        if len(commtype) == 1:
            kwargs.update(commtype[0])
            kwargs.setdefault('name', name)
            kwargs.setdefault('use_async', use_async)
            return new_comm(**kwargs)
        else:
            kwargs['comm_list'] = commtype
            commtype = 'fork'
    if (commtype is None) and kwargs.get('filetype', None):
        commtype = kwargs.pop('filetype')
    if commtype in ['ErrorComm', 'error']:
        kwargs['commtype'] = commtype
        commtype = kwargs['base_commtype']
    comm_cls = import_comm(commtype)
    if kwargs.get('is_interface', False):
        use_async = False
    async_kws = {}
    if use_async:
        from yggdrasil.communication.AsyncComm import AsyncComm
        async_kws = {k: kwargs.pop(k) for k in AsyncComm._async_kws
                     if k in kwargs}
    if use_async:
        kwargs['is_async'] = True
    out = comm_cls.new_comm(name, **kwargs)
    if use_async and (out._commtype not in [None, 'client',
                                            'server', 'fork']):
        from yggdrasil.communication.AsyncComm import AsyncComm
        out = AsyncComm(out, **async_kws)
    return out


def get_comm(name, **kwargs):
    r"""Return communicator for existing comm components.

    Args:
        name (str): Communicator name.
        **kwargs: Additional keyword arguments are passed to new_comm.

    Returns:
        Comm: Communicator of given class.

    """
    kwargs['dont_create'] = True
    return new_comm(name, **kwargs)
    

def cleanup_comms(commtype=None):
    r"""Call cleanup_comms for the appropriate communicator class.

    Args:
        commtype (str, optional): Name of communicator class. Defaults to
            'default' if not provided.

    Returns:
        int: Number of comms closed.

    """
    return import_comm(commtype).cleanup_comms()


@contextmanager
def open_file_comm(fname, mode, filetype='binary', **kwargs):
    r"""Context manager to open a file comm in a way similar to how
    Python opens file descriptors.

    Args:
        fname (str): Path to file that should be opened.
        mode (str): Mode that file should be opened in. Supported values
            include 'r', 'w', and 'a'.
        filetype (str, optional): Type of file being opened. Defaults to
            'binary'.

    Returns:
        CommBase: File comm.

    """
    comm = None
    try:
        comm_cls = import_component('file', filetype)
        if mode == 'r':
            kwargs['direction'] = 'recv'
        elif mode in ['w', 'a']:
            kwargs['direction'] = 'send'
            if mode == 'a':
                kwargs['append'] = True
        else:
            raise ValueError("Unsupported mode: '%s'" % mode)
        comm = comm_cls('file', address=fname, **kwargs)
        yield comm
    finally:
        if comm is not None:
            comm.close()


def get_open_fds():  # pragma: debug
    '''
    return the number of open file descriptors for current process
    
    .. warning: will only work on UNIX-like os-es.
    '''
    procs = subprocess.check_output(["lsof", '-w', "-p", str(os.getpid())])
    nprocs = len(procs.split(b'\n'))
    # nprocs = len(
    #     list(filter(
    #         lambda s: s and s[ 0 ] == b'f' and s[1: ].isdigit(),
    #         procs.split( b'\n' ) ) )
    # )
    return nprocs


__all__ = ['new_comm', 'get_comm', 'cleanup_comms']



# Import classes for input/output channels
from yggdrasil.interface.YggInterface import YggInput, YggOutput

# Initialize input/output channels
in_channel = YggInput('input')
out_channel = YggOutput('output')

# Loop until there is no longer input or the queues are closed
while True:

    # Receive input from input channel
    # If there is an error, the flag will be False
    flag, obj = in_channel.recv()
    if not flag:
        print("Python Model: No more input.")
        break

    # Print received message
    print('Python Model: %s' % str(obj))

    # Print keys
    print('Python Model: keys = %s' % str(obj.keys()))

    # Get floating point element
    co2 = obj['CO2']
    print('Python Model: CO2 = %f' % co2)

    # Get array element
    amaxtb = obj['AMAXTB']
    print('Python Model: AMAXTB = ')
    for x, y in zip(*amaxtb):
        print('\t%f\t%f' % (x, y))

    # Send output to output channel
    # If there is an error, the flag will be False
    flag = out_channel.send(obj)
    if not flag:
        raise RuntimeError("Python Model: Error sending output.")



import copy
import numpy as np
from yggdrasil import units, platform
from yggdrasil.metaschema.datatypes.tests import test_MetaschemaType as parent
from yggdrasil.metaschema.properties.ScalarMetaschemaProperties import (
    _valid_types)


class TestScalarMetaschemaType(parent.TestMetaschemaType):
    r"""Test class for ScalarMetaschemaType class with float."""
    _mod = 'ScalarMetaschemaType'
    _cls = 'ScalarMetaschemaType'
    _prec = 32
    _type = 'float'
    _shape = 1
    _array_contents = None

    @staticmethod
    def after_class_creation(cls):
        r"""Actions to be taken during class construction."""
        parent.TestMetaschemaType.after_class_creation(cls)
        if not cls._explicit:
            cls._typedef['subtype'] = cls._type
        if cls._type == 'bytes':
            dtype = 'S%d' % (cls._prec // 8)
        elif cls._type == 'unicode':
            dtype = 'U%d' % (cls._prec // 32)
        else:
            dtype = '%s%d' % (cls._type, cls._prec)
        if cls._array_contents is None:
            cls._array = np.ones(cls._shape, dtype)
        else:
            cls._array = np.array(cls._array_contents, dtype)
        if cls._type in ['bytes', 'unicode']:
            dtype_invalid = 'float'
        else:
            dtype_invalid = 'S10'
        cls._invalid_array = np.ones(cls._shape, dtype_invalid)
        if 'Array' not in cls._cls:
            cls._value = cls._array[0]
            cls._invalid_decoded.append(cls._array)
            cls._invalid_decoded.append(cls._invalid_array[0])
        else:
            cls._value = cls._array
            if cls._array.ndim == 1:
                cls._invalid_decoded.append(cls._array[0])
                cls._invalid_decoded.append(np.ones((3, 4), dtype))
            else:
                cls._invalid_decoded.append(cls._array[0][0])
                cls._invalid_decoded.append(cls._array[0])
            cls._invalid_decoded.append(cls._invalid_array)
        cls._valid_encoded = [{'type': cls.get_import_cls().name,
                               'precision': cls._prec,
                               'units': '',
                               'data': cls._value.tobytes()}]
        if not cls._explicit:
            cls._valid_encoded[0]['subtype'] = cls._type
        cls._valid_decoded = [cls._value]
        if cls._type == 'bytes':
            new_dtype = 'S%d' % (cls._prec * 2 // 8)
        elif cls._type == 'unicode':
            new_dtype = 'U%d' % (cls._prec * 2 // 32)
        else:
            new_dtype = '%s%d' % (cls._type, cls._prec * 2)
        if platform._is_win and (new_dtype == 'float128'):  # pragma: windows
            cls._prec_value = None
        else:
            prec_array = cls._array.astype(new_dtype)
            if 'Array' not in cls._cls:
                cls._prec_value = prec_array[0]
            else:
                cls._prec_value = prec_array
        cls._compatible_objects = [
            (cls._value, cls._value, None)]
        if cls._prec_value is not None:
            if not cls._explicit:
                cls._compatible_objects.append(
                    (cls._value, cls._prec_value, {'subtype': cls._type,
                                                   'precision': cls._prec * 2}))
            else:
                cls._compatible_objects.append(
                    (cls._value, cls._prec_value, {'precision': cls._prec * 2}))
        if 'Array' not in cls._cls:
            if cls._explicit:
                if cls._type == 'bytes':
                    cls._valid_normalize = [(1, b'1'),
                                            (u'1', b'1')]
                elif cls._type == 'unicode':
                    cls._valid_normalize = [(1, u'1'),
                                            (b'1', u'1')]
                else:
                    cls._valid_normalize = [(str(cls._value), cls._value),
                                            ('hello', 'hello')]
        if cls._explicit and ('Array' not in cls._cls):
            cls._invalid_encoded.append({'type': 'scalar',
                                         'subtype': 'invalid'})
        cls._invalid_validate.append(np.array([None, 1, list()],
                                              dtype=object))

    def test_from_array(self):
        r"""Test getting object from array."""
        test_val = self._value
        test_kws = {}
        if 'units' in self._typedef:
            test_val = units.add_units(test_val, self._typedef['units'])
            test_kws['unit_str'] = self._typedef['units']
        self.assert_equal(self.instance.from_array(self._array, **test_kws),
                          test_val)


# Dynamically create tests for dynamic and explicitly typed scalars
for t in _valid_types.keys():
    iattr_imp = {'_type': t}
    if t == 'complex':
        iattr_imp['_prec'] = 64
    elif t in ('bytes', 'unicode'):
        iattr_imp['_array_contents'] = ['one', 'two', 'three']
        max_len = len(max(iattr_imp['_array_contents'], key=len))
        if t == 'unicode':
            iattr_imp['_prec'] = max_len * 32
        else:
            iattr_imp['_prec'] = max_len * 8
    iattr_exp = copy.deepcopy(iattr_imp)
    iattr_exp['_cls'] = '%sMetaschemaType' % t.title()
    iattr_exp['_explicit'] = True
    if t == 'float':
        iattr_exp['_prec'] = 64
    cls_imp = type('TestScalarMetaschemaType_%s' % t,
                   (TestScalarMetaschemaType, ), iattr_imp)
    cls_exp = type('Test%s' % iattr_exp['_cls'],
                   (TestScalarMetaschemaType, ), iattr_exp)
    globals()[cls_imp.__name__] = cls_imp
    globals()[cls_exp.__name__] = cls_exp
    del cls_imp, cls_exp


class TestScalarMetaschemaType_prec(TestScalarMetaschemaType):
    r"""Test class for ScalarMetaschemaType class with precision."""

    @staticmethod
    def after_class_creation(cls):
        r"""Actions to be taken during class construction."""
        TestScalarMetaschemaType.after_class_creation(cls)
        cls._typedef['precision'] = cls._prec
        cls._valid_encoded.append(copy.deepcopy(cls._valid_encoded[0]))
        for x in cls._invalid_encoded:
            x['precision'] = cls._prec / 2  # compatible precision
        # Version with incorrect precision
        cls._invalid_encoded.append(copy.deepcopy(cls._valid_encoded[0]))
        if cls._prec_value is not None:
            cls._invalid_encoded[-1]['precision'] = cls._prec * 2
            cls._invalid_decoded.append(cls._prec_value)


class TestScalarMetaschemaType_units(TestScalarMetaschemaType):
    r"""Test class for ScalarMetaschemaType class with units."""

    @staticmethod
    def after_class_creation(cls):
        r"""Actions to be taken during class construction."""
        TestScalarMetaschemaType.after_class_creation(cls)
        cls._typedef['units'] = 'cm'
        cls._valid_encoded.append(copy.deepcopy(cls._valid_encoded[0]))
        cls._valid_encoded[-1]['units'] = 'cm'
        cls._valid_encoded.append(copy.deepcopy(cls._valid_encoded[0]))
        cls._valid_encoded[-1]['units'] = 'm'
        cls._valid_decoded.append(copy.deepcopy(cls._valid_decoded[0]))
        cls._valid_decoded[-1] = units.add_units(cls._valid_decoded[-1], 'm')
        # Version with incorrect units
        cls._invalid_encoded.append(copy.deepcopy(cls._valid_encoded[0]))
        cls._invalid_encoded[-1]['units'] = 's'



import re
import os
import numpy as np
import pandas as pd
import logging
from collections import OrderedDict
from yggdrasil import serialize, platform, constants
from yggdrasil.drivers.InterpretedModelDriver import InterpretedModelDriver
from yggdrasil.drivers.PythonModelDriver import PythonModelDriver
from yggdrasil.drivers.CModelDriver import CModelDriver
from yggdrasil.languages.R import install


logger = logging.getLogger(__name__)


class RModelDriver(InterpretedModelDriver):  # pragma: R
    r"""Base class for running R models.

    Args:
        name (str): Driver name.
        args (str or list): Argument(s) for running the model in matlab.
            Generally, this should be the full path to a Matlab script.
        **kwargs: Additional keyword arguments are passed to parent class's
            __init__ method.

    """
    _schema_subtype_description = ('Model is written in R.')
    language = 'R'
    language_aliases = ['r']
    language_ext = '.R'
    base_languages = ['python']
    default_interpreter = 'Rscript'
    # Dynamically setting the interface library causes circular a import so
    # it is defined here statically. For dynamic import, use the following:
    #     interface_library = PythonModelDriver.interface_library
    interface_library = 'yggdrasil'
    interface_dependencies = [x.split()[0] for x in
                              install.requirements_from_description()]
    # The Batch version causes output to saved to a file rather than directed to
    # stdout so use Rscript instead. For the batch version, use the following:
    #     default_interpreter_flags = ['CMD', 'BATCH' '--vanilla', '--silent']
    default_interpreter_flags = ['--default-packages=methods,utils']
    send_converters = {'table': serialize.consolidate_array}
    type_map = {
        'int': 'integer, bit64::integer64',
        'float': 'double',
        'string': 'character',
        'array': 'list',
        'object': 'list',
        'boolean': 'logical',
        'null': 'NA',
        'uint': 'integer',
        'complex': 'complex',
        'bytes': 'char (utf-8)',
        'unicode': 'char',
        '1darray': 'list',
        'ndarray': 'list',
        'ply': 'PlyDict',
        'obj': 'ObjDict',
        'schema': 'list'}
    function_param = {
        'import': 'source(\"{filename}\")',
        'istype': 'is({variable}, \"{type}\")',
        'len': 'length({variable})',
        'index': '{variable}[[{index}]]',
        'first_index': 1,
        'interface': 'library(yggdrasil)',
        'input': '{channel} <- YggInterface(\"YggInput\", \"{channel_name}\")',
        'output': '{channel} <- YggInterface(\"YggOutput\", \"{channel_name}\")',
        'python_interface': ('{channel} <- YggInterface(\"{python_interface}\", '
                             '\"{channel_name}\")'),
        'python_interface_format': ('{channel} <- YggInterface('
                                    '\"{python_interface}\", '
                                    '\"{channel_name}\", '
                                    '\"{format_str}\")'),
        'recv_function': '{channel}$recv',
        'send_function': '{channel}$send',
        'multiple_outputs': 'c({outputs})',
        'multiple_outputs_def': 'list({outputs})',
        'true': 'TRUE',
        'false': 'FALSE',
        'not': '!',
        'and': '&&',
        'comment': '#',
        'indent': 2 * ' ',
        'quote': '\"',
        'print_generic': 'print({object})',
        'print': 'print(\"{message}\")',
        'fprintf': 'print(sprintf(\"{message}\", {variables}))',
        'error': 'stop(\"{error_msg}\")',
        'block_end': '}',
        'if_begin': 'if ({cond}) {{',
        'if_elif': '}} else if ({cond}) {{',
        'if_else': '}} else {{',
        'for_begin': 'for ({iter_var} in {iter_begin}:{iter_end}) {{',
        'while_begin': 'while ({cond}) {{',
        'try_begin': 'tryCatch({',
        'try_except': '}}, error = function({error_var}) {{',
        'try_end': '})',
        'assign': '{name} <- {value}',
        'assign_mult': '{name} %<-% {value}',
        'function_def_begin': '{function_name} <- function({input_var}) {{',
        'return': 'return({output_var})',
        'function_def_regex': (
            r'{function_name} *(?:(?:\<-)|(?:=)) *function'
            r'\((?P<inputs>(?:.|(?:\r?\n))*?)\)\s*\{{'
            r'(?P<body>(?:.|(?:\r?\n))*?)'
            r'(?:return\((list\()?'
            r'(?P<outputs>(?:.|(?:\r?\n))*?)(?(3)\))\)'
            r'(?:.|(?:\r?\n))*?\}})'
            r'|(?:\}})'),
        'inputs_def_regex': r'\s*(?P<name>.+?)\s*(?:,|$)',
        'outputs_def_regex': r'\s*(?P<name>.+?)\s*(?:,|$)',
        'interface_regex': (
            r'(?P<indent>[ \t]*)'
            r'(?P<variable>[a-zA-Z\_]+[a-zA-Z\_0-9]*)\s*\<\-\s*'
            r'YggInterface\(\'(?P<class>[a-zA-Z\_]+[a-zA-Z\_0-9]*)\'\s*\,\s*'
            r'\'(?P<channel>[a-zA-Z\_]+[a-zA-Z\_0-9]*)\'\s*'
            r'(?P<args>[^\)]*)\)')}
    brackets = (r'{', r'}')
    zero_based = False

    @classmethod
    def is_library_installed(cls, lib, **kwargs):
        r"""Determine if a dependency is installed.

        Args:
            lib (str): Name of the library that should be checked.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            bool: True if the library is installed, False otherwise.

        """
        if lib not in cls._library_cache:
            try:
                cls.run_executable(['-e', 'library(%s)' % lib.split()[0]])
                cls._library_cache[lib] = True
            except RuntimeError as e:
                logger.info('Error checking for R library %s: %s' % (lib, e))
                cls._library_cache[lib] = False
        return cls._library_cache[lib]
        
    @classmethod
    def language_version(cls, **kwargs):
        r"""Determine the version of this language.

        Args:
            **kwargs: Keyword arguments are passed to the parent class's
                method.

        Returns:
            str: Version of compiler/interpreter for this language.

        """
        kwargs.setdefault('skip_interpreter_flags', True)
        return super(RModelDriver, cls).language_version(**kwargs)

    def run_model(self, *args, **kwargs):
        r"""Run the model. Unless overridden, the model will be run using
        run_executable.

        Args:
            *args: Arguments are passed to the parent class's method.
            **kwargs: Keyword arguments are passed to the parent class's
                method.

        """
        if self.with_valgrind:
            interp = kwargs.pop('interpreter', self.get_interpreter())
            interp_flags = kwargs.pop('interpreter_flags', [])
            if 'Rscript' in interp:
                interp = 'R'.join(interp.rsplit('Rscript', 1))
                interp_flags = []
                kwargs['skip_interpreter_flags'] = True
            interp_flags += [
                '--vanilla', '-d', 'valgrind %s'
                % ' '.join(self.valgrind_flags), '-f']
            kwargs['interpreter'] = interp
            kwargs['interpreter_flags'] = interp_flags
            kwargs['debug_flags'] = []
        return super(RModelDriver, self).run_model(*args, **kwargs)
        
    def set_env(self, **kwargs):
        r"""Get environment variables that should be set for the model process.

        Returns:
            dict: Environment variables for the model process.

        """
        out = super(RModelDriver, self).set_env(**kwargs)
        out['RETICULATE_PYTHON'] = PythonModelDriver.get_interpreter()
        if CModelDriver.is_language_installed():
            c_linker = CModelDriver.get_tool('linker')
            search_dirs = c_linker.get_search_path(env_only=True)
            out = CModelDriver.update_ld_library_path(out, paths_to_add=search_dirs,
                                                      add_to_front=True)
        return out
        
    @classmethod
    def comm_atexit(cls, comm):  # pragma: no cover
        r"""Operations performed on comm at exit including draining receive.
        
        Args:
            comm (CommBase): Communication object.

        """
        if comm.direction == 'recv':
            while comm.recv(timeout=0)[0]:
                comm.sleep()
        else:
            comm.send_eof()
            comm.linger()
        if not getattr(comm, 'dont_backlog', True):
            comm.linger_close()

    @classmethod
    def python2language(cls, pyobj):
        r"""Prepare a python object for transformation in R.

        Args:
            pyobj (object): Python object.

        Returns:
            object: Python object in a form that is R friendly.

        """
        logger.debug("python2language: %s, %s" % (pyobj, type(pyobj)))
        if isinstance(pyobj, tuple):
            return tuple([cls.python2language(x) for x in pyobj])
        elif isinstance(pyobj, list):
            return [cls.python2language(x) for x in pyobj]
        elif isinstance(pyobj, OrderedDict):
            return OrderedDict([(cls.python2language(k),
                                 cls.python2language(v))
                                for k, v in pyobj.items()])
        elif isinstance(pyobj, dict):
            return {cls.python2language(k): cls.python2language(v)
                    for k, v in pyobj.items()}
        elif isinstance(pyobj, np.string_):
            return pyobj.decode("utf-8")
        elif isinstance(pyobj, pd.DataFrame):
            # R dosn't have int64 and will cast 64bit ints as floats if passed
            # without casting them to int32 first
            for n in pyobj.columns:
                if pyobj[n].dtype == np.dtype('int64'):
                    pyobj[n] = pyobj[n].astype('int32')
        return pyobj
    
    @classmethod
    def write_model_wrapper(cls, model_file, model_function, **kwargs):
        r"""Return the lines required to wrap a model function as an integrated
        model.

        Args:
            model_file (str): Full path to the file containing the model
                function's declaration.
            model_function (str): Name of the model function.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            list: Lines of code wrapping the provided model with the necessary
                code to run it as part of an integration.

        """
        if platform._is_win and isinstance(model_file, str):  # pragma: windows
            model_file = model_file.replace('\\', '/')
        return super(RModelDriver, cls).write_model_wrapper(
            model_file, model_function, **kwargs)

    @classmethod
    def write_executable_import(cls, **kwargs):
        r"""Add import statements to executable lines.
       
        Args:
            **kwargs: Keyword arguments for import statement.

        Returns:
            list: Lines required to complete the import.
 
        """
        if kwargs.get('filename', None) and os.path.isfile(kwargs['filename']):
            with open(kwargs['filename'], 'r') as fd:
                contents = fd.read()
            # For functions that contain interface calls, split them out
            # and assign them to the global level
            if re.search(cls.function_param['interface_regex'], contents):
                new_contents = contents
                for m in re.finditer(cls.function_param['interface_regex'],
                                     contents):
                    mdict = m.groupdict()
                    assert('global_scope' in mdict['args'])
                    # if 'global_scope' not in mdict['args']:
                    #     mdict['args'] += ', global_scope=TRUE'
                    new_channel = (
                        '{indent}if (!exists("{variable}")) {{\n'
                        '{indent}  assign("{variable}", YggInterface('
                        '\'{class}\', \'{channel}\'{args}), '
                        'envir = .GlobalEnv)\n'
                        '{indent}}}').format(**mdict)
                    new_contents = new_contents.replace(m.group(0),
                                                        new_channel)
                out = new_contents.splitlines()
                return out
            if platform._is_win:  # pragma: windows
                kwargs['filename'] = kwargs['filename'].replace('\\', '/')
        return super(RModelDriver, cls).write_executable_import(**kwargs)

    # The following is only provided for the yggcc CLI for building R
    # packages and should not be used directly
    @classmethod
    def call_compiler(cls, package_dir, toolname=None, flags=None,
                      language='c++', verbose=False,
                      use_ccache=False):  # pragma: no cover
        r"""Build an R package w/ the yggdrasil compilers.

        Args:
            package_dir (str): Full path to the package directory.
            toolname (str, optional): Compilation tool that should be used.
                Defaults to None and the default tools will be used.
            flags (list, optional): Additional flags that should be passed
                to the build command. Defaults to [].
            language (str, optional): Language that the package is written in
                (e.g. c, fortran). Defautls to 'c++'.
            verbose (bool, optional): If True, information about the build
                process will be displayed. Defaults to False.
            use_ccache (bool, optional): If True, ccache will be added to
                the compilation executable. Defaults to False.

        """
        import shutil
        import subprocess
        from yggdrasil.components import import_component
        if flags is None:
            flags = []
        if isinstance(package_dir, list):
            assert(len(package_dir) == 1)
            package_dir = package_dir[0]
        cexec_vars = {'c': ['CC', 'CC_FOR_BUILD'],
                      'c++': ['CXX', 'CPP', 'CXX98', 'CXX11', 'CXX14',
                              'CXX17', 'CXX20', 'CXX_FOR_BUILD'],
                      'fortran': ['FC', 'F77']}
        cflag_vars = {'c': ['CFLAGS'],
                      'c++': ['CXXFLAGS', 'CXX98FLAGS', 'CXX11FLAGS',
                              'CXX14FLAGS', 'CXX17FLAGS', 'CXX20FLAGS'],
                      'fortran': ['FFFLAGS', 'FCFLAGS']}
        cstd_vars = {'c': [],
                     'c++': ['CXXSTD', 'CXX11STD', 'CXX14STD',
                             'CXX17STD', 'CXX20STD'],
                     'fortran': []}
        env = os.environ.copy()
        new_env = {}
        for x in constants.LANGUAGES['compiled']:
            # for x in [language]:
            drv = import_component('model', x)
            compiler = drv.get_tool('compiler', toolname=toolname)
            # archiver = compiler.archiver()
            env = drv.set_env_compiler(existing=env, compiler=compiler)
            cexec = compiler.get_executable(full_path=True)
            if use_ccache and shutil.which('ccache'):
                cexec = '%s %s' % ('ccache', cexec)
                env['CCACHE_NOHASHDIR'] = 'true'
            kws = {}
            if x == 'c++':
                kws['skip_standard_flag'] = True
            cflags0 = drv.get_compiler_flags(for_model=True, toolname=toolname,
                                             dry_run=True, compiler=compiler,
                                             dont_link=True, **kws)
            lflags = drv.get_linker_flags(for_model=True, toolname=toolname,
                                          dry_run=True, libtype='shared')
            # Remove flags that are unnecessary
            cflags = []
            stdflags = []
            for k in cflags0:
                if k.startswith("-std="):
                    stdflags.append(k)
                else:
                    cflags.append(k)
            for k in ['-c']:
                if k in cflags:
                    cflags.remove(k)
            for k in ['-shared']:
                if k in lflags:
                    lflags.remove(k)
            for k in cexec_vars[x]:
                env[k] = cexec
                new_env[k] = cexec
            for k in cflag_vars[x]:
                env[k] = ''
                new_env[k] = ' '.join(cflags)
            for k in cstd_vars[x]:
                env[k] = ''
                new_env[k] = ' '.join(stdflags)
            if language == x:
                env['LDFLAGS'] = ''
                new_env['LDFLAGS'] = ' '.join(lflags)
        cmd = ['R', 'CMD', 'INSTALL', '--no-html', '--no-help', '--no-docs',
               '--no-demo', '--no-multiarch', package_dir] + flags
        makevar = os.path.expanduser(os.path.join('~', '.R', 'Makevars'))
        makevar_copy = makevar + '_copy'
        try:
            if os.path.isfile(makevar):
                shutil.move(makevar, makevar_copy)
            with open(makevar, 'w') as fd:
                for k, v in new_env.items():
                    if k.startswith('PKG_'):
                        fd.write('%s+=%s\n' % (k, v))
                    else:
                        fd.write('%s=%s\n' % (k, v))
            if verbose:
                with open(makevar, 'r') as fd:
                    print(fd.read())
            subprocess.check_call(cmd, env=env)
        finally:
            os.remove(makevar)
            if os.path.isfile(makevar_copy):
                shutil.move(makevar_copy, makevar)



"""This modules offers various tools."""
from __future__ import print_function
import threading
import logging
import pprint
import os
import re
import sys
import sysconfig
try:
    from distutils import sysconfig as distutils_sysconfig
except ImportError:  # pragma: debug
    distutils_sysconfig = None
import warnings
import copy
import shutil
import inspect
import time
import signal
import uuid as uuid_gen
import subprocess
import importlib
import difflib
from yggdrasil import platform, constants
from yggdrasil.components import import_component, ComponentBase


logger = logging.getLogger(__name__)
YGG_MSG_EOF = b'EOF!!!'
YGG_MSG_BUF = 1024 * 2


_stack_in_log = False
_stack_in_timeout = False
if ((logging.getLogger("yggdrasil").getEffectiveLevel()
     <= logging.DEBUG)):  # pragma: debug
    _stack_in_log = False
    _stack_in_timeout = True


def apply_recurse(x, func, **kwargs):
    r"""Apply a function recursively to all elements of x if it is
    a list, tuple, or dictionary.

    Args:
        x (list, tuple, dict): Object to apply function to.
        func (function): Function to apply to elements of x.
        **kwargs: Additional keyword arguments are passed to each
            function call.

    Returns:
        object: Version of input, but after applying func.

    """
    if isinstance(x, (list, tuple)):
        out = [func(ix, **kwargs) for ix in x]
        if isinstance(x, tuple):
            out = tuple(out)
    elif isinstance(x, dict):
        out = {k: func(v, **kwargs) for k, v in x.items()}
    else:  # pragma: debug
        raise TypeError("Recursion not supported for type '%s'"
                        % type(x))
    return out


def bytes2str(x, recurse=False):
    r"""Convert bytes type to string type.

    Args:
        x (bytes): String.
        recurse (bool, optional): If True and x is a list, tuple, or
            dict, the coversion will recurse. Defaults to False.

    Returns:
        str: Decoded string version of x.

    """
    if isinstance(x, bytes):
        out = x.decode("utf-8")
    elif isinstance(x, str):
        out = str(x)
    elif isinstance(x, (list, tuple, dict)) and recurse:
        out = apply_recurse(x, bytes2str, recurse=True)
    else:  # pragma: debug
        raise TypeError("Cannot convert type '%s' to str." % type(x))
    return out


def str2bytes(x, recurse=False):
    r"""Convert string type to bytes type.

    Args:
        x (str): String.
        recurse (bool, optional): If True and x is a list, tuple, or
            dict, the coversion will recurse. Defaults to False.

    Returns:
        bytes: Encoded bytes version of x.

    """
    if isinstance(x, str):
        out = x.encode("utf-8")
    elif isinstance(x, bytes):
        out = bytes(x)
    elif isinstance(x, (list, tuple, dict)) and recurse:
        out = apply_recurse(x, str2bytes, recurse=True)
    else:  # pragma: debug
        raise TypeError("Cannot convert type '%s' to bytes." % type(x))
    return out


def add_line_numbers(lines, for_diff=False):
    r"""Add number to lines.

    Args:
        lines (list): Lines to number.
        for_diff (bool, optional): If True, those lines beginning with
            removed/notation diff characters (-, ?) will not be numbered.
            Defaults to False.

    Returns:
        list: Numbered lines.

    """
    out = []
    i = 0
    for line in lines:
        if for_diff and line.startswith(('-', '?')):
            out.append('    ' + line)
        else:
            i += 1
            out.append('%2d: %s' % (i, line))
    return out


def display_source(fname, number_lines=False, return_lines=False):
    r"""Display source code with syntax highlighting (if available).

    Args:
        fname (str, list): Full path(s) to one or more source files.
        number_lines (bool, optional): If True, line numbers will be added
            to the displayed examples. Defaults to False.
        return_lines (bool, optional): If True, the lines are returned rather
            than displayed. Defaults to False.

    """
    if isinstance(fname, list):
        out = ''
        for f in fname:
            iout = display_source(f, number_lines=number_lines,
                                  return_lines=return_lines)
            if return_lines:
                out += iout
        if return_lines:
            return out
        return
    if isinstance(fname, (bytes, str)):
        with open(fname, 'r') as fd:
            lines = fd.read()
        try:
            language = constants.EXT2LANG[os.path.splitext(fname)[-1]]
        except KeyError:
            language = None
        prefix = 'file: %s' % fname
    else:
        lines = inspect.getsource(fname)
        language = 'python'
        prefix = '%s: %s' % (type(fname), fname)
    try:
        from pygments import highlight
        from pygments.lexers import PythonLexer
        from pygments.lexers.data import YamlLexer
        from pygments.lexers.c_cpp import CLexer, CppLexer
        from pygments.lexers.r import SLexer
        from pygments.lexers.fortran import FortranLexer
        from pygments.lexers.matlab import MatlabLexer
        from pygments.lexers.html import XmlLexer
        from pygments.formatters import Terminal256Formatter
        lexer_map = {'python': PythonLexer,
                     'yaml': YamlLexer,
                     'c': CLexer,
                     'cxx': CppLexer,
                     'c++': CppLexer,
                     'r': SLexer,
                     'fortran': FortranLexer,
                     'matlab': MatlabLexer,
                     'xml': XmlLexer}
        lines = highlight(lines, lexer_map[language](),
                          Terminal256Formatter())
    except (ImportError, KeyError):
        pass
    if number_lines:
        lines = '\n'.join(add_line_numbers(lines.splitlines()))
    if return_lines:
        return lines
    lines = '%s\n%s\n%s\n' % (prefix, len(prefix) * '=', lines)
    print(lines)


def display_source_diff(fname1, fname2, number_lines=False,
                        return_lines=False):
    r"""Display a diff between two source code files with syntax highlighting
    (if available).

    Args:
        fname1 (str): Name of first source file.
        fname2 (src): Name of second source file.
        number_lines (bool, optional): If True, line numbers will be added
            to the displayed examples. Defaults to False.
        return_lines (bool, optional): If True, the lines are returned rather
            than displayed. Defaults to False.

    """
    src1 = display_source(fname1, return_lines=True)
    src2 = display_source(fname2, return_lines=True)
    diff = difflib.ndiff(src1.splitlines(), src2.splitlines())
    if number_lines:
        diff = add_line_numbers(diff, for_diff=True)
    if isinstance(fname1, str):
        prefix_type1 = 'file'
    else:
        prefix_type1 = str(type(fname1))
    if isinstance(fname2, str):
        prefix_type2 = 'file'
    else:
        prefix_type2 = str(type(fname2))
    prefix1 = '%s1: %s' % (prefix_type1, fname1)
    prefix2 = '%s2: %s' % (prefix_type2, fname2)
    lines = '%s\n%s\n%s\n%s\n' % (prefix1, prefix2,
                                  max(len(prefix1), len(prefix2)) * '=',
                                  '\n'.join(diff))
    if return_lines:
        return lines
    print(lines)
    

def get_fds():  # pragma: debug
    r"""Get a list of open file descriptors."""
    out = subprocess.check_output(
        'lsof -p {} | grep -v txt'.format(os.getpid()), shell=True)
    return out.splitlines()[1:]


def get_shell():
    r"""Get the type of shell that yggdrasil was called from.

    Returns:
        str: Name of the shell.

    """
    shell = os.environ.get('SHELL', None)
    if not shell:
        if platform._is_win:  # pragma: windows
            shell = os.environ.get('COMSPEC', None)
        else:
            shell = '/bin/sh'  # Default used by subprocess
        assert(shell)
    # return psutil.Process(os.getppid()).name()
    if platform._is_win:  # pragma: windows
        shell = shell.lower()
    return shell


def in_powershell():
    r"""Determine if yggdrasil is running from a Windows Powershell.

    Returns:
        bool: True if running from Powershell, False otherwise.

    """
    if not platform._is_win:
        return False
    shell = get_shell()
    return bool(re.match('pwsh|pwsh.exe|powershell.exe', shell))


def check_environ_bool(name, valid_values=['true', '1', True, 1]):
    r"""Check to see if a boolean environment variable is set to True.

    Args:
        name (str): Name of environment variable to check.
        valid_values (list, optional): Values for the environment variable
            that indicate it is True. These should all be lower case as
            the lower case version of the variable contents will be compared
            to the list. Defaults to ['true', '1'].

    Returns:
        bool: True if the environment variables is set and is one of the
            list valid_values (after being transformed to lower case).

    """
    return (os.environ.get(name, '').lower() in valid_values)


def get_python_c_library(allow_failure=False, libtype=None):
    r"""Determine the location of the Python C API library.

    Args:
        allow_failure (bool, optional): If True, the base name will be returned
            if the file cannot be located. Defaults to False.
        libtype (str, optional): Type of library that should be located.
            Valid values include 'static' and 'shared'. Defaults to 'shared'
            on Unix OSs and 'static' on Windows.

    Returns:
        str: Full path to the library.

    Raises:
        ValueError: If libtype is not 'static' or 'shared'.
        RuntimeError: If the base name for the library cannot be determined.
        RuntimeError: If the library cannot be located.

    """
    if libtype not in ['static', 'shared', None]:  # pragma: debug
        raise ValueError("libtype must be 'shared' or 'static', "
                         "'%s' not supported." % libtype)
    paths = sysconfig.get_paths()
    cvars = sysconfig.get_config_vars()
    if platform._is_win:  # pragma: windows
        libtype2ext = {'shared': '.dll', 'static': '.lib'}
        prefix = ''
        if libtype is None:
            libtype = 'shared'
        base = '%spython%s%s' % (prefix,
                                 cvars['py_version_nodot'],
                                 libtype2ext[libtype])
    elif sys.version_info[:2] < (3, 8):
        if libtype is None:
            libtype = 'shared'
        libtype2key = {'shared': 'LDLIBRARY', 'static': 'LIBRARY'}
        base = cvars.get(libtype2key[libtype], None)
    else:
        if libtype is None:
            libtype = 'shared'
        if platform._is_mac:
            libtype2ext = {'shared': '.dylib', 'static': '.a'}
        else:
            libtype2ext = {'shared': '.so', 'static': '.a'}
        prefix = 'lib'
        base = '%spython%s%s' % (prefix,
                                 cvars['py_version_short'],
                                 libtype2ext[libtype])
    if platform._is_mac and base.endswith('/Python'):  # pragma: no cover
        base = 'libpython%s.dylib' % cvars['py_version_short']
    if base is None:  # pragma: debug
        raise RuntimeError(("Could not determine base name for the Python "
                            "C API library.\n"
                            "sysconfig.get_paths():\n%s\n"
                            "sysconfig.get_config_vars():\n%s\n")
                           % (pprint.pformat(paths),
                              pprint.pformat(cvars)))
    dir_try = []
    for x in [get_conda_prefix(), cvars['prefix']]:
        if x:
            dir_try.append(x)
            if platform._is_win:  # pragma: windows
                dir_try.append(os.path.join(x, 'libs'))
            else:
                dir_try.append(os.path.join(x, 'lib'))
    for k in ["LIBPL", "LIBDIR", "LIBDEST", "Prefix", "ExecPrefix",
              "BaseExecPrefix"]:
        if cvars.get(k, None) and (cvars[k] not in dir_try):
            dir_try.append(cvars[k])
    for k in ['stdlib', 'purelib', 'platlib', 'platstdlib', 'data']:
        if paths.get(k, None) and (paths[k] not in dir_try):
            dir_try.append(paths[k])
    dir_try.append(os.path.join(paths['data'], 'lib'))
    if distutils_sysconfig is not None:
        dir_try.append(os.path.dirname(
            distutils_sysconfig.get_python_lib(True, True)))
    dir_try = set(dir_try)
    for idir in dir_try:
        x = os.path.join(idir, base)
        if os.path.isfile(x):
            return x
    if allow_failure:  # pragma: debug
        return base
    raise RuntimeError(("Could not determine the location of the Python "
                        "C API library: %s.\n"
                        "sysconfig.get_paths():\n%s\n"
                        "sysconfig.get_config_vars():\n%s\n")
                       % (base, pprint.pformat(paths),
                          pprint.pformat(cvars)))  # pragma: debug


def get_env_prefixes():
    r"""Determine the environment path prefix (virtualenv or conda) for
    the current environment.

    Returns:
        list: Full path to the directory prefixes used for the current
            environments if one (or more) exists. If neither a
            virtualenv or conda prefix can be located, None is returned.

    Raises:
        RuntimeError: If both virtualenv and conda environments are
            located.

    """
    out = []
    venv = get_venv_prefix()
    cenv = get_conda_prefix()
    if venv:
        out.append(venv)
    if cenv:
        out.append(cenv)
    return out


def get_venv_prefix():
    r"""Determine the virtualenv path prefix for the current environment.

    Returns:
        str: Full path to the directory prefix used for the current
            virtualenv environment if one exists. If virtualenv cannot
            be located, None is returned.

    """
    return os.environ.get('VIRTUAL_ENV', None)


def get_conda_prefix():
    r"""Determine the conda path prefix for the current environment.

    Returns:
        str: Full path to the directory prefix used for the current conda
            environment if one exists. If conda cannot be located, None is
            returned.

    """
    conda_prefix = os.environ.get('CONDA_PREFIX', None)
    # This part should be enabled if the conda base enviroment dosn't have
    # CONDA_PREFIX set. Older version of conda behaved this way so it is
    # possible that a future release will as well.
    # if not conda_prefix:
    #     conda_prefix = shutil.which('conda')
    #     if conda_prefix is not None:
    #         conda_prefix = os.path.dirname(os.path.dirname(conda_prefix))
    return conda_prefix


def get_conda_env():
    r"""Determine the name of the current conda environment.

    Returns:
        str: Name of the current conda environment if one is activated. If a
            conda environment is not activated, None is returned.

    """
    return os.environ.get('CONDA_DEFAULT_ENV', None)


def get_subprocess_language():
    r"""Determine the language of the calling process.

    Returns:
        str: Name of the programming language responsible for the subprocess.

    """
    return os.environ.get('YGG_MODEL_LANGUAGE', 'python')


def get_subprocess_language_driver():
    r"""Determine the driver for the langauge of the calling process.

    Returns:
        ModelDriver: Class used to handle running a model of the process language.

    """
    return import_component('model', get_subprocess_language())


def is_subprocess():
    r"""Determine if the current process is a subprocess.

    Returns:
        bool: True if YGG_SUBPROCESS environment variable is True, False
            otherwise.

    """
    return check_environ_bool('YGG_SUBPROCESS')


def find_all(name, path, verification_func=None):
    r"""Find all instances of a file with a given name within the directory
    tree starting at a given path.

    Args:
        name (str): Name of the file to be found (with the extension).
        path (str, None): Directory where search should start. If set to
            None on Windows, the current directory and PATH variable are
            searched.
        verification_func (function, optional): Function that returns
            True when a file is valid and should be returned and False
            otherwise. Defaults to None and is ignored.

    Returns:
        list: All instances of the specified file.

    """
    result = []
    try:
        if platform._is_win:  # pragma: windows
            if path is None:
                out = subprocess.check_output(["where", name],
                                              env=os.environ,
                                              stderr=subprocess.STDOUT)
            else:
                out = subprocess.check_output(["where", "/r", path, name],
                                              env=os.environ,
                                              stderr=subprocess.STDOUT)
        else:
            args = ["find", "-L", path, "-type", "f", "-name", name]
            pfind = subprocess.Popen(args, env=os.environ,
                                     stderr=subprocess.PIPE,
                                     stdout=subprocess.PIPE)
            (stdoutdata, stderrdata) = pfind.communicate()
            out = stdoutdata
            for line in stderrdata.splitlines():
                if b'Permission denied' not in line:
                    raise subprocess.CalledProcessError(pfind.returncode,
                                                        ' '.join(args),
                                                        output=stderrdata)
    except subprocess.CalledProcessError:
        out = ''
    if not out.isspace():
        result = sorted(out.splitlines())
    result = [os.path.normcase(os.path.normpath(bytes2str(m)))
              for m in result]
    if verification_func is not None:
        result = [x for x in result if verification_func(x)]
    return result


def locate_file(fname, environment_variable='PATH', directory_list=None,
                show_alternates=False, verification_func=None):
    r"""Locate a file within a set of paths defined by a list or environment
    variable.

    Args:
        fname (str, list): One or more possible names of the file that should be
            located. If a list is provided, the path for the first entry for
            which a match could be located will be returned and subsequent entries
            will not be checked.
        environment_variable (str): Environment variable containing the set of
            paths that should be searched. Defaults to 'PATH'. If None, this
            keyword argument will be ignored. If a list is provided, it is
            assumed to be a list of environment variables that should be
            searched in the specified order.
        directory_list (list): List of paths that should be searched in addition
            to those specified by environment_variable. Defaults to None and is
            ignored. These directories will be searched be for those in the
            specified environment variables.
        show_alternates (bool, optional): If True and there is more
            than one match, the alternate matches will be printed in
            a warning message. Defaults to False.
        verification_func (function, optional): Function that returns
            True when a file is valid and should be returned and False
            otherwise. Defaults to None and is ignored.

    Returns:
        bool, str: Full path to the located file if it was located, False
            otherwise.

    """
    if isinstance(fname, list):
        out = False
        for ifname in fname:
            out = locate_file(ifname, environment_variable=environment_variable,
                              directory_list=directory_list,
                              show_alternates=show_alternates,
                              verification_func=verification_func)
            if out:
                break
        return out
    out = []
    if ((platform._is_win and (environment_variable == 'PATH')
         and (directory_list is None))):  # pragma: windows
        out += find_all(fname, None,
                        verification_func=verification_func)
    else:
        if directory_list is None:
            directory_list = []
        if environment_variable is not None:
            if not isinstance(environment_variable, list):
                environment_variable = [environment_variable]
            for x in environment_variable:
                directory_list += os.environ.get(x, '').split(os.pathsep)
        for path in directory_list:
            if path:
                out += find_all(fname, path,
                                verification_func=verification_func)
            if out and (not show_alternates):
                break
    if not out:
        return False
    first = out[0]
    if show_alternates:  # pragma: debug
        out = set(out)
        out.remove(first)
        if len(out) > 0:
            warnings.warn(("More than one (%d) match to %s:\n%s\n "
                           + "Using first match (%s)") %
                          (len(out) + 1, fname, pprint.pformat(out),
                           first), RuntimeWarning)
    return first


def locate_path(fname, basedir=os.path.abspath(os.sep)):
    r"""Find the full path to a file using where on Windows."""
    try:
        if platform._is_win:  # pragma: windows
            out = subprocess.check_output(["dir", fname, "/s/b"], shell=True,
                                          cwd=basedir)
            # out = subprocess.check_output(["where", fname])
        else:
            # find . -name "filetofind" 2>&1 | grep -v 'permission denied'
            out = subprocess.check_output(["find", basedir, "-name", fname])  # ,
            # "2>&1", "|", "grep", "-v", "'permission denied'"])
            # out = subprocess.check_output(["locate", "-b", "--regex",
            #                                "^%s" % fname])
    except subprocess.CalledProcessError:  # pragma: debug
        return False
    if out.isspace():  # pragma: debug
        return False
    out = bytes2str(out).splitlines()
    return out


def remove_path(fpath, timer_class=None, timeout=None):
    r"""Delete a single file.

    Args:
        fpath (str): Full path to a file or directory that should be
            removed.
        timer_class (YggClass, optional): Class that should be used to
            generate a timer that is used to wait for file to be removed.
            Defaults to None and a new class instance will be created.
        timeout (float, optional): Time (in seconds) that should be
            waited before raising an error that a file cannot be removed.
            Defaults to None and will be set by the timer_class.

    Raises:
        RuntimeError: If the product cannot be removed.

    """
    if timer_class is None:
        timer_class = YggClass()
    if os.path.isdir(fpath):
        T = timer_class.start_timeout(t=timeout)
        while ((not T.is_out) and os.path.isdir(fpath)):
            try:
                shutil.rmtree(fpath)
            except BaseException:  # pragma: debug
                if os.path.isdir(fpath):
                    timer_class.sleep()
                if T.is_out:
                    raise
        timer_class.stop_timeout()
        if os.path.isdir(fpath):  # pragma: debug
            raise RuntimeError("Failed to remove directory: %s" % fpath)
    elif os.path.isfile(fpath):
        T = timer_class.start_timeout(t=timeout)
        while ((not T.is_out) and os.path.isfile(fpath)):
            try:
                os.remove(fpath)
            except BaseException:  # pragma: debug
                if os.path.isfile(fpath):
                    timer_class.sleep()
                if T.is_out:
                    raise
        timer_class.stop_timeout()
        if os.path.isfile(fpath):  # pragma: debug
            raise RuntimeError("Failed to remove file: %s" % fpath)


def get_supported_platforms():
    r"""Get a list of the platforms supported by yggdrasil.

    Returns:
        list: The name of platforms supported by yggdrasil.

    """
    return copy.deepcopy(platform._supported_platforms)


def get_supported_lang():
    r"""Get a list of the model programming languages that are supported
    by yggdrasil.

    Returns:
        list: The names of programming languages supported by yggdrasil.
    
    """
    # from yggdrasil import schema
    # s = schema.get_schema()
    # out = s['model'].subtypes
    # if 'r' in out:
    #     out[out.index('r')] = 'R'
    out = constants.LANGUAGES['all'].copy()
    if 'c++' in out:
        out[out.index('c++')] = 'cpp'
    return list(set(out))


def get_supported_type():
    r"""Get a list of the data types that are supported by yggdrasil.

    Returns:
        list: The names of data types supported by yggdrasil.

    """
    from yggdrasil.metaschema.datatypes import get_registered_types
    return list(get_registered_types().keys())


def get_supported_comm(dont_include_value=False):
    r"""Get a list of the communication mechanisms supported by yggdrasil.

    Args:
        dont_include_value (bool, optional): If True, don't include the
            ValueComm in the list returned. Defaults to False.

    Returns:
        list: The names of communication mechanisms supported by yggdrasil.

    """
    from yggdrasil import schema
    excl_list = ['CommBase', 'DefaultComm', 'default']
    if dont_include_value:
        excl_list += ['ValueComm', 'value']
    s = schema.get_schema()
    out = s['comm'].subtypes
    for k in excl_list:
        if k in out:
            out.remove(k)
    return list(set(out))


def is_lang_installed(lang):
    r"""Check to see if yggdrasil can run models written in a programming
    language on the current machine.

    Args:
        lang (str): Programming language to check.

    Returns:
        bool: True if models in the provided language can be run on the current
            machine, False otherwise.

    """
    drv = import_component('model', lang)
    return drv.is_installed()


def is_comm_installed(comm, language=None):
    r"""Check to see if yggdrasil can use a communication mechanism on the
    current machine.

    Args:
        comm (str): Communication mechanism to check.
        language (str, optional): Specific programming language that
            communication mechanism should be check for. Defaults to None and
            all supported languages will be checked.

    Returns:
        bool: True if the communication mechanism can be used on the current
            machine, False otherwise.

    """
    cmm = import_component('comm', comm)
    return cmm.is_installed(language=language)


def get_installed_lang():
    r"""Get a list of the languages that are supported by yggdrasil on the
    current machine. This checks for the necessary interpreters, licenses, and/or
    compilers.

    Returns:
        list: The name of languages supported on the current machine.
    
    """
    out = []
    all_lang = get_supported_lang()
    for k in all_lang:
        if is_lang_installed(k):
            out.append(k)
    return out


def get_installed_comm(language=None, dont_include_value=False):
    r"""Get a list of the communication channel types that are supported by
    yggdrasil on the current machine. This checks the operating system,
    supporting libraries, and broker credentials. The order indicates the
    prefered order of use.

    Args:
        language (str, optional): Specific programming language that installed
            comms should be located for. Defaults to None and all languages
            supported on the current platform will be checked.
        dont_include_value (bool, optional): If True, don't include the
            ValueComm in the list returned. Defaults to False.


    Returns:
        list: The names of the the communication channel types supported on
            the current machine.

    """
    out = []
    all_comm = get_supported_comm(dont_include_value=dont_include_value)
    for k in all_comm:
        if is_comm_installed(k, language=language):
            out.append(k)
    # Fix order to denote preference
    out_sorted = []
    for k in ['zmq', 'ipc', 'rmq']:
        if k in out:
            out.remove(k)
            out_sorted.append(k)
    out_sorted += out
    return out_sorted


def get_default_comm():
    r"""Get the default comm that should be used for message passing."""
    if 'YGG_DEFAULT_COMM' in os.environ:
        _default_comm = os.environ['YGG_DEFAULT_COMM']
        # if not is_comm_installed(_default_comm, language='any'):  # pragma: debug
        #     raise Exception('Unsupported default comm %s set by YGG_DEFAULT_COMM' % (
        #                     _default_comm))
    else:
        comm_list = get_installed_comm()
        if len(comm_list) > 0:
            _default_comm = comm_list[0]
        else:  # pragma: windows
            # Locate comm that maximizes languages that can be run
            tally = {}
            for c in get_supported_comm(dont_include_value=True):
                tally[c] = 0
                for lang in get_supported_lang():
                    if is_comm_installed(c, language=lang):
                        tally[c] += 1
            _default_comm = max(tally)
            if tally[_default_comm] == 0:  # pragma: debug
                raise Exception('Could not locate an installed comm.')
    if _default_comm.endswith('Comm'):
        _default_comm = import_component('comm', _default_comm)._commtype
    if _default_comm == 'rmq':  # pragma: debug
        raise NotImplementedError('RMQ cannot be the default comm because '
                                  + 'there is not an RMQ C interface.')
    return _default_comm


def get_YGG_MSG_MAX(comm_type=None):
    r"""Get the maximum message size for a given comm type.

    Args:
        comm_type (str, optional): The name of the communication type that the
            maximum message size should be returned for. Defaults to result of
            get_default_comm() if not provided.

    Returns:
        int: Maximum message size (in bytes).

    """
    if comm_type is None:
        comm_type = get_default_comm()
    if comm_type in ['ipc', 'IPCComm']:
        # OS X limit is 2kb
        out = 1024 * 2
    else:
        out = 2**20
    return out


# https://stackoverflow.com/questions/35772001/
# how-to-handle-the-signal-in-python-on-windows-machine
def kill(pid, signum):
    r"""Kill process by mapping signal number.

    Args:
        pid (int): Process ID.
        signum (int): Signal that should be sent.

    """
    if platform._is_win:  # pragma: debug
        sigmap = {signal.SIGINT: signal.CTRL_C_EVENT,
                  signal.SIGBREAK: signal.CTRL_BREAK_EVENT}
        if signum in sigmap and pid == os.getpid():
            # we don't know if the current process is a
            # process group leader, so just broadcast
            # to all processes attached to this console.
            pid = 0
        thread = threading.current_thread()
        handler = signal.getsignal(signum)
        # work around the synchronization problem when calling
        # kill from the main thread.
        if (((signum in sigmap) and (thread.name == 'MainThread')
             and callable(handler) and ((pid == os.getpid()) or (pid == 0)))):
            event = threading.Event()

            def handler_set_event(signum, frame):
                event.set()
                return handler(signum, frame)

            signal.signal(signum, handler_set_event)
            try:
                print("calling interrupt", pid)
                os.kill(pid, sigmap[signum])
                # busy wait because we can't block in the main
                # thread, else the signal handler can't execute.
                while not event.is_set():
                    pass
                print("after interrupt")
            finally:
                signal.signal(signum, handler)
                print("in finally")
        else:
            os.kill(pid, sigmap.get(signum, signum))
    else:
        os.kill(pid, signum)


def sleep(interval):
    r"""Sleep for a specified number of seconds.

    Args:
        interval (float): Time in seconds that process should sleep.

    """
    time.sleep(interval)


def safe_eval(statement, **kwargs):
    r"""Run eval with a limited set of builtins and Python libraries/functions.

    Args:
        statement (str): Statement that should be evaluated.
        **kwargs: Additional keyword arguments are variables that are made available
            to the statement during evaluation.

    Returns:
        object: Result of the eval.

    """
    safe_dict = {}
    _safe_lists = {'math': ['acos', 'asin', 'atan', 'atan2', 'ceil', 'cos',
                            'cosh', 'degrees', 'e', 'exp', 'fabs', 'floor', 'fmod',
                            'frexp', 'hypot', 'ldexp', 'log', 'log10', 'modf', 'pi',
                            'pow', 'radians', 'sin', 'sinh', 'sqrt', 'tan', 'tanh'],
                   'builtins': ['abs', 'any', 'bool', 'bytes', 'float', 'int', 'len',
                                'list', 'map', 'max', 'min', 'repr', 'set', 'str',
                                'sum', 'tuple', 'type'],
                   'numpy': ['array', 'int8', 'int16', 'int32', 'int64',
                             'uint8', 'uint16', 'uint32', 'uint64',
                             'float16', 'float32', 'float64'],
                   'yggdrasil.units': ['get_data', 'add_units'],
                   'unyt.array': ['unyt_quantity', 'unyt_array']}
    for mod_name, func_list in _safe_lists.items():
        mod = importlib.import_module(mod_name)
        for func in func_list:
            safe_dict[func] = getattr(mod, func)
    safe_dict.update(kwargs)
    # The following replaces <Class Name(a, b)> style reprs with calls to classes
    # identified in self._no_eval_class
    # regex = r'<([^<>]+)\(([^\(\)]+)\)>'
    # while True:
    #     match = re.search(regex, statement)
    #     if not match:
    #         break
    #     cls_repl = self._no_eval_class.get(match.group(1), False)
    #     if not cls_repl:
    #         raise ValueError("Expression '%s' in '%s' is not eval friendly."
    #                          % (match.group(0), statement))
    #     statement = statement.replace(match.group(0),
    #                                   '%s(%s)' % (cls_repl, match.group(2)), 1)
    return eval(statement, {"__builtins__": None}, safe_dict)


def eval_kwarg(x):
    r"""If x is a string, eval it. Otherwise just return it.

    Args:
        x (str, obj): String to be evaluated as an object or an object.

    Returns:
        obj: Result of evaluated string or the input object.

    """
    if isinstance(x, str):
        try:
            return eval(x)
        except NameError:
            return x
    return x


class ProxyMeta(type):
    r"""Metaclass for handling proxy."""

    _special_names = [
        '__abs__', '__add__', '__and__', '__call__', '__cmp__', '__coerce__',
        '__contains__', '__delitem__', '__delslice__', '__div__', '__divmod__',
        '__eq__', '__float__', '__floordiv__', '__ge__', '__getitem__',
        '__getslice__', '__gt__', '__hex__', '__iadd__', '__iand__',
        '__idiv__', '__idivmod__', '__ifloordiv__', '__ilshift__', '__imod__',
        '__imul__', '__int__', '__invert__', '__ior__', '__ipow__', '__irshift__',
        '__isub__', '__iter__', '__itruediv__', '__ixor__', '__le__', '__len__',
        '__long__', '__lshift__', '__lt__', '__mod__', '__mul__', '__ne__',
        '__neg__', '__oct__', '__or__', '__pos__', '__pow__', '__radd__',
        '__rand__', '__rdiv__', '__rdivmod__', '__reduce__', '__reduce_ex__',
        '__repr__', '__reversed__', '__rfloorfiv__', '__rlshift__', '__rmod__',
        '__rmul__', '__ror__', '__rpow__', '__rrshift__', '__rshift__', '__rsub__',
        '__rtruediv__', '__rxor__', '__setitem__', '__setslice__', '__sub__',
        '__truediv__', '__xor__', 'next', '__hash__'
    ]
    
    def __new__(cls, classname, bases, attrs):
        overrides = attrs.get('__overrides__', [])
        overrides.extend(attrs.get('__slots__', []))
        overrides.extend(k for k in attrs.keys() if k not in
                         ['__overrides__'])
        for base in bases:
            overrides.extend(getattr(base, '__overrides__', []))
        assert('_wrapped' in overrides)
        attrs['__overrides__'] = overrides
        
        def make_method(name):
            def method(self, *args, **kwargs):
                mtd = getattr(object.__getattribute__(self, "_wrapped"), name)
                return mtd(*args, **kwargs)
            return method

        for name in cls._special_names:
            if name not in overrides:
                attrs[name] = make_method(name)
        return type.__new__(cls, classname, bases, attrs)


class ProxyObject(metaclass=ProxyMeta):
    r"""Proxy for another object."""
    # http://code.activestate.com/recipes/496741-object-proxying/
    
    __slots__ = ["_wrapped", "__weakref__"]
    
    def __init__(self, wrapped):
        object.__setattr__(self, "_wrapped", wrapped)

    def __getattribute__(self, name):
        if name in object.__getattribute__(self, '__overrides__'):
            return object.__getattribute__(self, name)
        return getattr(object.__getattribute__(self, "_wrapped"), name)
    
    def __delattr__(self, name):
        if name in object.__getattribute__(self, '__overrides__'):
            object.__delattr__(self, name)
            return
        delattr(object.__getattribute__(self, "_wrapped"), name)
        
    def __setattr__(self, name, value):
        if name in object.__getattribute__(self, '__overrides__'):
            object.__setattr__(self, name, value)
            return
        setattr(object.__getattribute__(self, "_wrapped"), name, value)

    def __reduce__(self):
        return (object.__getattribute__(self, "__class__"),
                (object.__getattribute__(self, "_wrapped"), ))

    def __reduce_ex__(self, proto):
        return object.__getattribute__(self, "__reduce__")()
    
    # Special cases
    def __bool__(self):
        return bool(object.__getattribute__(self, "_wrapped"))
    
    def __str__(self):
        return str(object.__getattribute__(self, "_wrapped"))
    
    def __bytes__(self):
        return bytes(object.__getattribute__(self, "_wrapped"))
    
    def __repr__(self):
        return repr(object.__getattribute__(self, "_wrapped"))


class YggPopen(subprocess.Popen):
    r"""Uses Popen to open a process without a buffer. If not already set,
    the keyword arguments 'bufsize', 'stdout', and 'stderr' are set to
    0, subprocess.PIPE, and subprocess.STDOUT respectively. This sets the
    output stream to unbuffered and directs both stdout and stderr to the
    stdout pipe. In addition this class overrides Popen.kill() to allow
    processes to be killed with CTRL_BREAK_EVENT on windows.

    Args:
        args (list, str): Shell command or list of arguments that should be
            run.
        forward_signals (bool, optional): If True, flags will be set such
            that signals received by the spawning process will be forwarded
            to the child process. If False, the signals will not be forwarded.
            Defaults to True.
        **kwargs: Additional keywords arguments are passed to Popen.

    """
    def __init__(self, cmd_args, forward_signals=True, for_matlab=False, **kwargs):
        # stdbuf only for linux
        if platform._is_linux:
            stdbuf_args = ['stdbuf', '-o0', '-e0']
            if isinstance(cmd_args, str):
                cmd_args = ' '.join(stdbuf_args + [cmd_args])
            else:
                cmd_args = stdbuf_args + cmd_args
        kwargs.setdefault('bufsize', 0)
        kwargs.setdefault('stdout', subprocess.PIPE)
        kwargs.setdefault('stderr', subprocess.STDOUT)
        # To prevent forward of signals, process will have a new process group
        if not forward_signals:
            if platform._is_win:  # pragma: windows
                # TODO: Make sure that Matlab handled correctly since pty not
                # guaranteed on windows
                kwargs.setdefault('preexec_fn', None)
                kwargs.setdefault('creationflags',
                                  subprocess.CREATE_NEW_PROCESS_GROUP)
            else:
                if for_matlab:  # pragma: matlab
                    import pty
                    # Matlab requires a tty so a pty is used here to allow
                    # the process to be lanched in a new process group.
                    # Related Materials:
                    # - https://www.mathworks.com/matlabcentral/answers/
                    #       359992-system-call-bizarre-behavior
                    # - https://gist.github.com/thepaul/1206753
                    # - https://stackoverflow.com/questions/30139401/
                    #       filter-out-command-that-needs-a-terminal-in-python-
                    #       subprocess-module
                    master_fd, slave_fd = pty.openpty()
                    kwargs.setdefault('stdin', slave_fd)

                kwargs.setdefault('preexec_fn', os.setpgrp)
        # if platform._is_win:  # pragma: windows
        #     kwargs.setdefault('universal_newlines', True)
        super(YggPopen, self).__init__(cmd_args, **kwargs)

    def kill(self, *args, **kwargs):
        r"""On windows using CTRL_BREAK_EVENT to kill the process."""
        if platform._is_win:  # pragma: windows
            self.send_signal(signal.CTRL_BREAK_EVENT)
        else:
            super(YggPopen, self).kill(*args, **kwargs)


def popen_nobuffer(*args, **kwargs):
    r"""Uses Popen to open a process without a buffer. If not already set,
    the keyword arguments 'bufsize', 'stdout', and 'stderr' are set to
    0, subprocess.PIPE, and subprocess.STDOUT respectively. This sets the
    output stream to unbuffered and directs both stdout and stderr to the
    stdout pipe.

    Args:
        args (list, str): Shell command or list of arguments that should be
            run.
        forward_signals (bool, optional): If True, flags will be set such
            that signals received by the spawning process will be forwarded
            to the child process. If False, the signals will not be forwarded.
            Defaults to True.
        **kwargs: Additional keywords arguments are passed to Popen.

    Returns:
        YggPopen: Process that was started.

    """
    return YggPopen(*args, **kwargs)


def pprint_encoded(obj, *args, **kwargs):
    r"""Pretty print an object, catching encoding errors as necessary.

    Args:
        obj (object): Python object to pprint.
        *args: Additional arguments are passed to pprint.pformat.
        **kwargs: Additional keyword arguments are passed to pprint.pformat.

    """
    print_encoded(pprint.pformat(obj, *args, **kwargs))


def print_encoded(msg, *args, **kwargs):
    r"""Print bytes to stdout, encoding if possible.

    Args:
        msg (str, bytes): Message to print.
        *args: Additional arguments are passed to print.
        **kwargs: Additional keyword arguments are passed to print.


    """
    if not isinstance(msg, (str, bytes)):
        msg = str(msg)
    try:
        print(bytes2str(msg), *args, **kwargs)
    except (UnicodeEncodeError, UnicodeDecodeError):  # pragma: debug
        logger.debug("sys.stdout.encoding = %s, cannot print unicode",
                     sys.stdout.encoding)
        kwargs.pop('end', None)
        try:
            print(msg, *args, **kwargs)
        except UnicodeEncodeError:  # pragma: debug
            print(str2bytes(msg), *args, **kwargs)


class TimeOut(object):
    r"""Class for checking if a period of time has been elapsed.

    Args:
        max_time (float, bbol): Maximum period of time that should elapse before
            'is_out' returns True. If False, 'is_out' will never return True.
            Providing 0 indicates that 'is_out' should immediately return True.
        key (str, optional): Key that was used to register the timeout. Defaults
            to None.

    Attributes:
        max_time (float): Maximum period of time that should elapsed before
            'is_out' returns True.
        start_time (float): Result of time.time() at start.
        key (str): Key that was used to register the timeout.

    """

    def __init__(self, max_time, key=None):
        self.max_time = max_time
        self.start_time = time.perf_counter()
        self.key = key
        self.checked = False

    @property
    def elapsed(self):
        r"""float: Total time that has elapsed since the start."""
        return time.perf_counter() - self.start_time
    
    @property
    def is_out(self):
        r"""bool: True if there is not any time remaining. False otherwise."""
        if self.max_time is False:
            return False
        if (not self.checked) and (self.max_time == 0):
            out = False
        else:
            out = (self.elapsed > self.max_time)
        self.checked = True
        return out


# def single_use_method(func):
#     r"""Decorator for marking functions that should only be called once."""
#     def wrapper(*args, **kwargs):
#         if getattr(func, '_single_use_method_called', False):
#             logger.info("METHOD %s ALREADY CALLED" % func)
#             return
#         else:
#             func._single_use_method_called = True
#             return func(*args, **kwargs)
#     return wrapper


class YggLoggerAdapter(logging.LoggerAdapter):
    r"""Logger adapter for use with YggClass."""

    def __init__(self, class_name, instance_name, *args, **kwargs):
        self._class_name = class_name
        self._instance_name = instance_name
        super(YggLoggerAdapter, self).__init__(*args, **kwargs)
    
    def process(self, msg, kwargs):
        r"""Process logging message."""
        if _stack_in_log:  # pragma: no cover
            stack = inspect.stack()
            the_class = os.path.splitext(os.path.basename(
                stack[2][0].f_globals["__file__"]))[0]
            the_line = stack[2][2]
            the_func = stack[2][3]
            prefix = '%s(%s).%s[%d]' % (the_class,
                                        self._instance_name,
                                        the_func, the_line)
        else:
            prefix = '%s(%s)' % (self._class_name,
                                 self._instance_name)
        new_msg = '%s: %s' % (prefix, self.as_str(msg))
        return new_msg, kwargs

    def as_str(self, obj):
        r"""Return str version of object if it is not already a string.

        Args:
            obj (object): Object that should be turned into a string.

        Returns:
            str: String version of provided object.

        """
        if not isinstance(obj, str):
            obj_str = str(obj)
        else:
            obj_str = obj
        return obj_str


class YggClass(ComponentBase):
    r"""Base class for Ygg classes.

    Args:
        name (str): Name used for component in log messages.
        uuid (str, optional): Unique ID for this instance. Defaults to None
            and is assigned.
        working_dir (str, optional): Working directory. If not provided, the
            current working directory is used.
        timeout (float, optional): Maximum time (in seconds) that should be
            spent waiting on a process. Defaults to 60.
        sleeptime (float, optional): Time that class should sleep for when
            sleep is called. Defaults to 0.01.
        **kwargs: Additional keyword arguments are passed to the ComponentBase
            initializer.

    Attributes:
        name (str): Class name.
        uuid (str): Unique ID for this instance.
        sleeptime (float): Time that class should sleep for when sleep called.
        longsleep (float): Time that the class will sleep for when waiting for
            longer tasks to complete (10x longer than sleeptime).
        timeout (float): Maximum time that should be spent waiting on a process.
        working_dir (str): Working directory.
        errors (list): List of errors.
        sched_out (obj): Output from the last scheduled task with output.
        logger (logging.Logger): Logger object for this object.
        suppress_special_debug (bool): If True, special_debug log messages
            are suppressed.

    """

    _base_defaults = ['name', 'uuid', 'working_dir', 'timeout', 'sleeptime']

    def __init__(self, name=None, uuid=None, working_dir=None,
                 timeout=60.0, sleeptime=0.01, **kwargs):
        # Defaults
        if name is None:
            name = ''
        if uuid is None:
            uuid = str(uuid_gen.uuid4())
        if working_dir is None:
            working_dir = os.getcwd()
        # Assign attributes
        self._name = name
        self.uuid = uuid
        self.sleeptime = sleeptime
        self.longsleep = self.sleeptime * 10
        self.timeout = timeout
        self._timeouts = {}
        self.working_dir = working_dir
        self.errors = []
        self.sched_out = None
        self.suppress_special_debug = False
        self._periodic_logs = {}
        self._old_loglevel = None
        self._old_encoding = None
        self.debug_flag = False
        # Call super class, adding in schema properties
        for k in self._base_defaults:
            if k in self._schema_properties:
                kwargs[k] = getattr(self, k)
        super(YggClass, self).__init__(**kwargs)
        self.logger = YggLoggerAdapter(
            self.__class__.__name__, self.print_name,
            logging.getLogger(self.__module__), {})

    def __getstate__(self):
        state = super(YggClass, self).__getstate__()
        del state['logger']
        # thread_attr = {}
        for k, v in list(state.items()):
            if isinstance(v, (threading._CRLock, threading._RLock,
                              threading.Event, threading.Thread)):  # pragma: debug
                self.warning("Special treatment of threading objects "
                             "currently disabled.")
            # if isinstance(v, (threading._CRLock, threading._RLock)):
            #     thread_attr.setdefault('threading.RLock', [])
            #     thread_attr['threading.RLock'].append((k, (), {}))
            # elif isinstance(v, threading.Event):
            #     thread_attr.setdefault('threading.Event', [])
            #     thread_attr['threading.Event'].append((k, (), {}))
            # elif isinstance(v, threading.Thread):
            #     assert(not v.is_alive())
            #     attr = {'name': v._name, 'group': None,
            #             'daemon': v.daemon, 'target': v._target,
            #             'args': v._args, 'kwargs': v._kwargs}
            #     thread_attr.setdefault('threading.Thread', [])
            #     thread_attr['threading.Thread'].append((k, (), attr))
        # for attr_list in thread_attr.values():
        #     for k in attr_list:
        #         state.pop(k[0])
        # state['thread_attr'] = thread_attr
        return state

    def __setstate__(self, state):
        super(YggClass, self).__setstate__(state)
        self.logger = YggLoggerAdapter(
            self.__class__.__name__, self.print_name,
            logging.getLogger(self.__module__), {})

    def __deepcopy__(self, memo):
        r"""Don't deep copy since threads cannot be copied."""
        return self

    @property
    def name(self):
        r"""str: Name of the class object."""
        return self._name

    @property
    def print_name(self):
        r"""str: Name of the class object."""
        return self._name.replace('%', '%%')

    def language_info(self, languages):
        r"""Only do info debug message if the language is one of those specified."""
        if not isinstance(languages, (list, tuple)):
            languages = [languages]
        languages = [lang.lower() for lang in languages]
        if get_subprocess_language().lower() in languages:  # pragma: debug
            return self.logger.info
        else:
            return self.dummy_log

    @property
    def interface_info(self):
        r"""Only do info debug message if is interface."""
        if is_subprocess():  # pragma: debug
            return self.logger.info
        else:
            return self.dummy_log

    def debug_log(self):  # pragma: debug
        r"""Turn on debugging."""
        self.info("Setting debug_log")
        from yggdrasil.config import get_ygg_loglevel, set_ygg_loglevel
        self._old_loglevel = get_ygg_loglevel()
        set_ygg_loglevel('DEBUG')

    def reset_log(self):  # pragma: debug
        r"""Resetting logging to prior value."""
        from yggdrasil.config import set_ygg_loglevel
        if self._old_loglevel is not None:
            set_ygg_loglevel(self._old_loglevel)

    def pprint(self, obj, block_indent=0, indent_str='    ', **kwargs):
        r"""Use pprint to represent an object as a string.

        Args:
            obj (object): Python object to represent.
            block_indent (int, optional): Number of indents that should be
                placed in front of the entire block. Defaults to 0.
            indent_str (str, optional): String that should be used to indent.
                Defaults to 4 spaces.
            **kwargs: Additional keyword arguments are passed to pprint.pformat.

        Returns:
            str: String representation of object using pprint.

        """
        sblock = block_indent * indent_str
        out = sblock + pprint.pformat(obj, **kwargs).replace('\n', '\n' + sblock)
        return out

    def display(self, msg='', *args, **kwargs):
        r"""Print a message, no log."""
        msg, kwargs = self.logger.process(msg, kwargs)
        print(msg % args)

    def verbose_debug(self, *args, **kwargs):
        r"""Log a verbose debug level message."""
        return self.logger.log(9, *args, **kwargs)
        
    def dummy_log(self, *args, **kwargs):
        r"""Dummy log function that dosn't do anything."""
        pass

    def periodic_debug(self, key, period=10):
        r"""Log that should occur periodically rather than with every call.

        Arguments:
            key (str): Key that should be used to identify the debug message.
            period (int, optional): Period (in number of messages) that messages
                should be logged at. Defaults to 10.

        Returns:
            method: Logging method to be used.

        """
        if key in self._periodic_logs:
            self._periodic_logs[key] += 1
        else:
            self._periodic_logs[key] = 0
        if (self._periodic_logs[key] % period) == 0:
            return self.logger.debug
        else:
            return self.dummy_log

    @property
    def special_debug(self):
        r"""Log debug level message contingent of supression flag."""
        if not self.suppress_special_debug:
            return self.logger.debug
        else:
            return self.dummy_log

    @property
    def info(self):
        r"""Log an info level message."""
        return self.logger.info

    @property
    def debug(self):
        r"""Log a debug level message."""
        return self.logger.debug

    @property
    def critical(self):
        r"""Log a critical level message."""
        return self.logger.critical

    @property
    def warn(self):
        r"""Log a warning level message."""
        return self.logger.warning

    @property
    def warning(self):
        r"""Log a warning level message."""
        return self.logger.warning

    @property
    def error(self):
        r"""Log an error level message."""
        self.errors.append('ERROR')
        return self.logger.error
        # return super(YggClass, self).error

    @property
    def exception(self):
        r"""Log an exception level message."""
        exc_info = sys.exc_info()
        if exc_info is not None and exc_info != (None, None, None):
            self.errors.append('ERROR')
            return self.logger.exception
            # return super(YggClass, self).exception
        else:
            return self.error

    def print_encoded(self, msg, *args, **kwargs):
        r"""Print bytes to stdout, encoding if possible.

        Args:
            msg (str, bytes): Message to print.
            *args: Additional arguments are passed to print.
            **kwargs: Additional keyword arguments are passed to print.


        """
        return print_encoded(msg, *args, **kwargs)

    def printStatus(self):
        r"""Print the class status."""
        self.logger.info('%s(%s): ', self.__module__, self.print_name)

    def _task_with_output(self, func, *args, **kwargs):
        self.sched_out = func(*args, **kwargs)

    def sched_task(self, t, func, args=None, kwargs=None, store_output=False,
                   name=None):
        r"""Schedule a task that will be executed after a certain time has
        elapsed.

        Args:
            t (float): Number of seconds that should be waited before task
                is executed.
            func (object): Function that should be executed.
            args (list, optional): Arguments for the provided function.
                Defaults to [].
            kwargs (dict, optional): Keyword arguments for the provided
                function. Defaults to {}.
            store_output (bool, optional): If True, the output from the
                scheduled task is stored in self.sched_out. Otherwise, it is not
                stored. Defaults to False.
            name (str, optional): Name for the task.

        Returns:
            threading.Timer: The timer object.

        """
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        self.sched_out = None
        if store_output:
            args = [func] + args
            func = self._task_with_output
        tobj = threading.Timer(t, func, args=args, kwargs=kwargs)
        if name is not None:
            tobj.name = name
        tobj.start()
        return tobj

    def sleep(self, t=None):
        r"""Have the class sleep for some period of time.

        Args:
            t (float, optional): Time that class should sleep for. If not
                provided, the attribute 'sleeptime' is used.

        """
        if t is None:
            t = self.sleeptime
        sleep(t)

    @property
    def timeout_key(self):  # pragma: no cover
        r"""str: Key identifying calling object and method."""
        return self.get_timeout_key()

    def get_timeout_key(self, key_level=0, key_suffix=None):
        r"""Return a key for a given level in the stack, relative to the
        function calling get_timeout_key.

        Args:
            key_level (int, optional): Positive integer indicating the level of
                the calling class and function/method that should be used to
                key the timeout. 0 is the class and function/method that is 2
                steps higher in the stack. Higher values use classes and
                function/methods further up in the stack. Defaults to 0.
            key_suffix (str, optional): String that should be appended to the
                end of the generated key. Defaults to None and is ignored.

        Returns:
            str: Key identifying calling object and method.

        """
        if _stack_in_timeout:  # pragma: debug
            stack = inspect.stack()
            fcn = stack[key_level + 2][3]
            cls = os.path.splitext(os.path.basename(stack[key_level + 2][1]))[0]
            key = '%s(%s).%s.%s' % (cls, self.print_name, fcn,
                                    threading.current_thread().name)
        else:
            key = '%s(%s).%s' % (str(self.__class__).split("'")[1], self.print_name,
                                 threading.current_thread().name)
        if key_suffix is not None:
            key += key_suffix
        return key

    def start_timeout(self, t=None, key=None, key_level=0, key_suffix=None):
        r"""Start a timeout for the calling function/method.

        Args:
            t (float, optional): Maximum time that the calling function should
                wait before timeing out. If not provided, the attribute
                'timeout' is used.
            key (str, optional): Key that should be associated with the timeout
                that is created. Defaults to None and is set by the calling
                class and function/method (See `get_timeout_key`).
            key_level (int, optional): Positive integer indicating the level of
                the calling class and function/method that should be used to
                key the timeout. 0 is the class and function/method that called
                start_timeout. Higher values use classes and function/methods
                further up in the stack. Defaults to 0.
            key_suffix (str, optional): String that should be appended to the
                end of the generated key. Defaults to None and is ignored.

        Raises:
            KeyError: If the key already exists.

        """
        if t is None:
            t = self.timeout
        if key is None:
            key = self.get_timeout_key(key_level=key_level, key_suffix=key_suffix)
        if key in self._timeouts:
            raise KeyError("Timeout already registered for %s" % key)
        self._timeouts[key] = TimeOut(t, key=key)
        return self._timeouts[key]

    def check_timeout(self, key=None, key_level=0):
        r"""Check timeout for the calling function/method.

        Args:
            key (str, optional): Key for timeout that should be checked.
                Defaults to None and is set by the calling class and
                function/method (See `timeout_key`).
            key_level (int, optional): Positive integer indicating the level of
                the calling class and function/method that should be used to
                key the timeout. 0 is the class and function/method that called
                start_timeout. Higher values use classes and function/methods
                further up in the stack. Defaults to 0.

        Raises:
            KeyError: If there is not a timeout registered for the specified
                key.

        """
        if key is None:
            key = self.get_timeout_key(key_level=key_level)
        if key not in self._timeouts:
            raise KeyError("No timeout registered for %s" % key)
        t = self._timeouts[key]
        return t.is_out
        
    def stop_timeout(self, key=None, key_level=0, key_suffix=None, quiet=False):
        r"""Stop a timeout for the calling function method.

        Args:
            key (str, optional): Key for timeout that should be stopped.
                Defaults to None and is set by the calling class and
                function/method (See `timeout_key`).
            key_level (int, optional): Positive integer indicating the level of
                the calling class and function/method that should be used to
                key the timeout. 0 is the class and function/method that called
                start_timeout. Higher values use classes and function/methods
                further up in the stack. Defaults to 0.
            key_suffix (str, optional): String that should be appended to the
                end of the generated key. Defaults to None and is ignored.
            quiet (bool, optional): If True, error message on timeout exceeded
                will be debug log. Defaults to False.

        Raises:
            KeyError: If there is not a timeout registered for the specified
                key.

        """
        if key is None:
            key = self.get_timeout_key(key_level=key_level, key_suffix=key_suffix)
        if key not in self._timeouts:
            raise KeyError("No timeout registered for %s" % key)
        t = self._timeouts[key]
        if t.is_out and t.max_time > 0:
            if quiet:
                self.debug("Timeout for %s at %5.2f/%5.2f s" % (
                    key, t.elapsed, t.max_time))
            else:
                self.info("Timeout for %s at %5.2f/%5.2f s" % (
                    key, t.elapsed, t.max_time))
        del self._timeouts[key]



import os
from yggdrasil.examples.tests.test_types import TestExampleTypes


def dst(received_data):
    typename = os.environ.get('TEST_TYPENAME', None)
    assert(typename is not None)
    TestExampleTypes.check_received_data(typename, received_data)
    print('CONFIRMED: %s' % typename)
    return
